import sys
import os
from PyQt6.QtWidgets import QApplication
from vehicle_gui.main_widget import VehicleGUI
from vehicle_gui import VEHICLE_DIR
from vehicle_gui.vcl_bindings import CACHE_DIR
from vehicle_gui.resource_view.property_view import RENDERERS_DIR
import signal

signal.signal(signal.SIGINT, signal.SIG_DFL)

os.makedirs(VEHICLE_DIR, exist_ok=True)
os.makedirs(CACHE_DIR, exist_ok=True)
os.makedirs(RENDERERS_DIR, exist_ok=True)

def main():
    app = QApplication(sys.argv)
    # Set application style
    app.setStyle("Fusion")
    
    try:
        editor = VehicleGUI()
        editor.show()
        exit_code = app.exec()
        
        # Use os._exit to avoid Python cleanup that causes segfaults
        os._exit(exit_code)
        
    except Exception as e:
        print(f"Error during application execution: {e}")
        import traceback
        traceback.print_exc()
        os._exit(1)

if __name__ == "__main__":
    main()


"""
Module edge_graphics.py

This module contains the class EdgeGraphics and its concrete children
DirectEdgeGraphics and BezierEdgeGraphics for displaying edges connecting
the blocks

Original author: Andrea Gimelli, Giacomo Rosato, Stefano Demarchi

"""

import abc
import math

from PyQt6.QtCore import Qt, QRectF, QPointF
from PyQt6.QtGui import QPen, QColor, QPainter, QPolygonF, QPainterPath
from PyQt6 import QtCore, QtGui
from PyQt6.QtWidgets import QGraphicsPathItem, QStyleOptionGraphicsItem

import vehicle_gui.query_view.styling.dimension as dim
import vehicle_gui.query_view.styling.palette as palette
from ..base_types import Edge, SocketType


class EdgeGraphics(QGraphicsPathItem):
    """Graphics representation of an Edge domain model"""

    def __init__(self, edge: 'Edge', parent=None):
        super().__init__(parent)

        # Reference to edge domain model
        self.edge_ref = edge

        if self.edge_ref.view_dim:
            self._pen = QPen(QColor(palette.DARK_TEAL))
        else:
            self._pen = QPen(QColor(palette.DARK_ORANGE))
        self._pen.setWidth(2)
        self.setZValue(-1)

        # Edge dimension label
        self.label = ''

        # Position
        self.src_pos = [0, 0]
        self.dest_pos = [200, 200]

        self.update()

    @abc.abstractmethod
    def update_path(self):
        """
        Abstract method to be implemented

        """

        pass

    @abc.abstractmethod
    def calc_path(self) -> QPainterPath:
        """
        Abstract method to be implemented

        """

        pass

    def set_label(self, text):
        self.label = text
        self.update()

    def build_arrow(self) -> list:
        """
        This method computes and returns the three vertices of the arrow

        Returns
        ----------
        list
            Three QPointF objects

        """

        radius = 7

        xs, ys = self.src_pos
        xd, yd = self.dest_pos[0] - radius, self.dest_pos[1]

        arrow_dimension = 5

        # Point1: equals self.position_destination
        point1 = QPointF(xd, yd)

        # Angular coefficient given 2 points
        try:
            ang_cf = (yd - ys) / (xd - xs)  # If xd-xs == 0 raises exception

            # Find the angle in degrees: atan(ang_cf)
            theta = math.atan(ang_cf)

            # Find point (coordinate) distant arrow_dimension pixel from point destination on the line
            p_auxiliary = QPointF(xd - arrow_dimension * math.cos(theta), yd - arrow_dimension * math.sin(theta))
            gamma = math.atan(-1 / ang_cf)

            # Point2
            point2 = QPointF(p_auxiliary.x() - arrow_dimension * math.cos(gamma),
                             p_auxiliary.y() - arrow_dimension * math.sin(gamma))

            # Point3
            point3 = QPointF(p_auxiliary.x() + arrow_dimension * math.cos(gamma),
                             p_auxiliary.y() + arrow_dimension * math.sin(gamma))

        except ZeroDivisionError:
            p_auxiliary = QPointF(xd - arrow_dimension, yd)
            point2 = QPointF(p_auxiliary.x(), p_auxiliary.y() - arrow_dimension)
            point3 = QPointF(p_auxiliary.x(), p_auxiliary.y() + arrow_dimension)

        return [point1, point2, point3]

    def paint(self, painter: QPainter, option: 'QStyleOptionGraphicsItem', widget=None) -> None:
        """
        Draw the edge, the arrow in the destination and the label

        """

        # Draw edge path
        self.update_path()
        painter.setPen(self._pen)
        painter.setBrush(Qt.BrushStyle.NoBrush)
        painter.drawPath(self.path())

        # Draw edge label
        painter.setBrush(QColor('yellow'))
        label_rect = QRectF((self.src_pos[0] + self.dest_pos[0]) / 2 - 48,
                            (self.src_pos[1] + self.dest_pos[1]) / 2, 100, 40)
        painter.drawText(label_rect, Qt.AlignmentFlag.AlignCenter, self.label)

        # No arrow - just solid lines

    def shape(self) -> 'QtGui.QPainterPath':
        return self.calc_path()

    def boundingRect(self) -> 'QtCore.QRectF':
        return self.shape().boundingRect()


class DirectEdgeGraphics(EdgeGraphics):
    """
    This class displays a direct edge from point A to point B

    """

    def update_path(self):
        """
        This method draws the actual line between two positions

        """

        path = QPainterPath(QPointF(self.src_pos[0], self.src_pos[1]))
        path.lineTo(self.dest_pos[0], self.dest_pos[1])
        self.setPath(path)

    def boundingRect(self) -> QRectF:
        return self.shape().boundingRect()

    def shape(self) -> QPainterPath:
        return self.path()

    def paint(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget=None) -> None:
        painter.setBrush(QtCore.Qt.BrushStyle.NoBrush)
        painter.setPen(self._pen)
        painter.drawPath(self.path())


class BezierEdgeGraphics(EdgeGraphics):
    """Graphics representation using a bezier path for the edge"""

    def __init__(self, edge, parent=None):
        super().__init__(edge, parent)
        super().__init__(edge, parent)

    def update_path(self):
        """Update path for hierarchical vertical connections with symmetric curves"""
        vertical_distance = abs(self.dest_pos[1] - self.src_pos[1])
        control_point_offset = min(vertical_distance * 0.4, 100)
        
        path = QPainterPath(QPointF(self.src_pos[0], self.src_pos[1]))
        path.cubicTo(
            self.src_pos[0], self.src_pos[1] + control_point_offset,   
            self.dest_pos[0], self.dest_pos[1] - control_point_offset, 
            self.dest_pos[0], self.dest_pos[1]
        )
        self.setPath(path)

    def calc_path(self) -> QPainterPath:
        """
        Compute the cubic Bezier line connection for hierarchical vertical layout

        Returns
        ----------
        QPainterPath
            The path of this edge with symmetric vertical curves

        """
        s = self.src_pos
        d = self.dest_pos
        
        vertical_distance = abs(d[1] - s[1])
        control_point_offset = min(vertical_distance * 0.4, 100) 
        
        cpx_s = 0  
        cpy_s = control_point_offset  
        cpx_d = 0  
        cpy_d = -control_point_offset  

        path = QPainterPath(QPointF(s[0], s[1]))
        path.cubicTo(s[0] + cpx_s, s[1] + cpy_s, d[0] + cpx_d, d[1] + cpy_d,
                     d[0], d[1])

        return path


import idx2numpy
import os
import time
from PyQt6.QtWidgets import (
    QWidget, QVBoxLayout, QComboBox, QLabel, QTextEdit, QStackedLayout,
    QPushButton, QFileDialog, QHBoxLayout, QSizePolicy
)
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QIcon
from typing import List, Type, Dict
from collections import OrderedDict

from vehicle_gui.vcl_bindings import CACHE_DIR
from vehicle_gui.counter_example_view.base_renderer import *


def _wait_until_stable_size(path: str, checks: int = 3, interval: float = 0.05) -> bool:
    """
    Return True if the file's size is stable over a series of checks.
    """
    try:
        last = os.path.getsize(path)
        if last <= 4:
            return False
        for _ in range(checks - 1):
            time.sleep(interval)
            now = os.path.getsize(path)
            if now != last:
                last = now
        time.sleep(interval)
        return os.path.getsize(path) == last and last > 4
    except FileNotFoundError:
        return False


def decode_counter_examples(cache_dir: str = CACHE_DIR) -> dict:
    """Decode counterexamples from IDX files in assignment directories."""
    subdirs = [
        d for d in os.listdir(cache_dir)
        if os.path.isdir(os.path.join(cache_dir, d)) and d.endswith("-assignments")
    ]

    counter_examples = {}
    for subdir in subdirs:
        subdir_path = os.path.join(cache_dir, subdir)
        for filename in os.listdir(subdir_path):
            full_path = os.path.join(subdir_path, filename)
            if not os.path.isfile(full_path):
                continue

            if not _wait_until_stable_size(full_path, checks=4, interval=0.05):
                continue

            var_name = filename.strip('\"')
            key = f"{subdir}-{var_name}"
            try:
                tensors = idx2numpy.convert_from_file(full_path)
                counter_examples[key] = tensors
            except Exception as e:
                print(f"Error decoding {full_path}: {e}")

    return counter_examples


class CounterExampleWidget(QWidget):
    """Widget for displaying individual counterexamples with navigation."""

    def __init__(self, parent=None):
        """Initialize the counterexample widget. Modes is a dict of variable names to lists of renderers supported for that variable."""
        super().__init__(parent)
        self.data_map = {}
        self.renderers = {}
        self.var_index = {}
        self.ce_paths = []
        self.ce_current_index = 0
        self.parent_ref = parent

        # Navigation controls
        self.name_label = QLabel("No data loaded")
        self.name_label.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.prev_button = QPushButton("◀")
        self.prev_button.setFixedSize(32, 32)

        self.next_button = QPushButton("▶")
        self.next_button.setFixedSize(32, 32)

        nav_layout = QHBoxLayout()
        nav_layout.addWidget(self.prev_button)
        nav_layout.addWidget(self.name_label)
        nav_layout.addWidget(self.next_button)

        self.stack = QStackedLayout()

        # Main layout
        main_layout = QVBoxLayout()
        main_layout.addLayout(nav_layout)
        
        # Create a container widget for the stack
        stack_container = QWidget()
        stack_container.setLayout(self.stack)
        main_layout.addWidget(stack_container)
        
        self.setLayout(main_layout)

        # Connect signals
        self.prev_button.clicked.connect(self._go_previous)
        self.next_button.clicked.connect(self._go_next)

    def set_modes(self, modes: Dict[str, BaseRenderer]):
        """Set the rendering modes."""
        # Ignore if no new modes provided
        if len(modes) == 0:
            return
        
        # Rebuild stack
        while self.stack.count():
            widget = self.stack.takeAt(0).widget()
            if widget:
                widget.setParent(None)

        self.var_index = {}
        self.renderers = {}

        ind = 0
        for var_name, renderer in modes.items():
            self.stack.addWidget(renderer.widget)
            self.var_index[var_name] = ind
            self.renderers[var_name] = renderer
            ind += 1
        
        # If we have data and just got renderers, update the display
        if len(modes) > 0 and self.ce_paths:
            self.update_display()

    def set_data(self, data: dict):
        """Set the counterexample data."""
        self.data_map = data
        self.ce_paths = list(data.keys())
        self.ce_current_index = 0
        if self.ce_paths:
            self._update_current_names()
        
        # Only update display if we have data, otherwise wait for renderers
        if self.ce_paths and len(self.renderers) > 0:
            self.update_display()
        elif not self.ce_paths:
            self.update_display()  # Clear display when no data

    def update_display(self):
        """Update the display based on current data and mode."""
        if not self.ce_paths:
            self.name_label.setText("No data")
            self.prev_button.hide()
            self.next_button.hide()
            return

        self.prev_button.show()
        self.next_button.show()

        key = self.ce_paths[self.ce_current_index]
        content = self.data_map.get(key, [])
        self.name_label.setText(f"{key}")

        # Render the data for the current variable if it has a renderer
        compound_key = f"{self.current_prop_name}-{self.current_var_name}"
        
        if compound_key in self.renderers:
            renderer = self.renderers[compound_key]
            try:
                renderer.render(content)
                self.stack.setCurrentIndex(self.var_index[compound_key])
            except Exception as e:
                print(f"Error during rendering: {e}")
        else:
            # No renderer available for this variable
            print(f"No renderer available for {compound_key}")

    def _go_previous(self):
        """Navigate to previous counterexample."""
        if self.ce_current_index > 0:
            self.ce_current_index -= 1
        else:
            self.ce_current_index = len(self.ce_paths) - 1

        # Update property and variable names
        self._update_current_names()
        self.update_display()

    def _go_next(self):
        """Navigate to next counterexample."""
        if self.ce_current_index < len(self.ce_paths) - 1:
            self.ce_current_index += 1
        else:
            self.ce_current_index = 0
        
        # Update property and variable names
        self._update_current_names()
        self.update_display()

    def _update_current_names(self):
        """Update current property and variable names based on current counterexample."""
        key = self.ce_paths[self.ce_current_index]
        key_parts = key.split('-')

        # Find the assignments part and extract property and variable
        assignments_idx = key_parts.index('assignments')
        self.current_prop_name = '-'.join(key_parts[:assignments_idx])
        self.current_var_name = '-'.join(key_parts[assignments_idx + 1:])


class CounterExampleTab(QWidget):
    """Main tab widget for counterexample visualization."""
    def __init__(self, parent=None):
        super().__init__(parent)
        self.layout = QVBoxLayout()

        # Control layout
        control_layout = QHBoxLayout()

        # Folder display
        self.folder_label = QLabel(CACHE_DIR)
        counter_examples_json = decode_counter_examples(CACHE_DIR)
        control_layout.addWidget(self.folder_label)

        # Content widget
        self.content_widget = CounterExampleWidget(parent=self)
        self.content_widget.set_data(counter_examples_json)
        self.layout.addWidget(self.content_widget)
        self.setLayout(self.layout)

    def set_modes(self, modes: Dict[str, BaseRenderer]):
        """Set the rendering modes for variables."""
        self.content_widget.set_modes(modes)

    def refresh_from_cache(self):
        """Re-read counter examples from the cache directory."""
        print(f"=== refresh_from_cache called ===")
        if os.path.exists(CACHE_DIR):
            counter_examples_json = decode_counter_examples(CACHE_DIR)
            print(f"Decoded {len(counter_examples_json)} counter examples")
            self.content_widget.set_data(counter_examples_json)
        else:
            print(f"Cache directory {CACHE_DIR} does not exist")



import os
import json

from PyQt6.QtWidgets import QTabWidget, QTextEdit, QTabBar, QSizePolicy, QWidget, QVBoxLayout, QComboBox, QLabel, QHBoxLayout, QSplitter
from PyQt6.QtCore import pyqtSignal, Qt
from pathlib import Path

from vehicle_gui.query_view.verification.blocks import Status, PropertyQuantifier
from vehicle_gui.query_view.graphics.block_graphics import BlockGraphics
from vehicle_gui.query_view.graphics.graphics_scene import GraphicsScene
from vehicle_gui.query_view.graphics.graphics_view import GraphicsView
from vehicle_gui.query_view.verification import VerificationWorkflow
from vehicle_gui.vcl_bindings import CACHE_DIR


class PropertyLoader:
    """Handles loading and parsing of VCL properties"""
    
    def __init__(self, workflow_generator):
        self.workflow_generator = workflow_generator
        self.global_query_id = 0
        self.property_name = None
    
    def load_property(self, property_name: str, cache_location: Path):
        """Load a property and create its workflow"""
        self.global_query_id = 0
        self.workflow_generator.clear_workflow()
        self.property_name = property_name
        plan_path = cache_location / f"{property_name}.vcl-plan"
        if plan_path.exists():
            self._create_workflow_from_plan(str(plan_path), property_name)
        else:
            raise FileNotFoundError(f"Property plan file not found: {plan_path}")
    
    def _create_workflow_from_plan(self, plan_path: str, title: str):
        """Create workflow from VCL plan file using pure JSON structure"""
        try:
            with open(plan_path, 'r') as f:
                plan_data = json.load(f)
            prop = self.workflow_generator.add_property(title=title)

            # Extract the root structure from the plan
            query_meta = plan_data.get('queryMetaData', {}).get('contents', {})
            
            # Handle flat structure with single disjunction of Queries
            if query_meta.get('tag') == 'Query':
                contents = query_meta.get('contents', {})
                queries = contents.get('queries', {}).get('unDisjunctAll', [])
                or_block = self.workflow_generator.add_or(prop)
                negated = contents.get('negated', False)
                for _ in queries:
                    self.global_query_id += 1
                    query_path = os.path.join(CACHE_DIR, f"{self.property_name}-query{self.global_query_id}.txt")
                    self.workflow_generator.add_query(self.global_query_id, or_block, query_path, is_negated=negated)

            # Handle multi-level structure
            else:
                self._parse_node(prop, query_meta)
                
            self.workflow_generator._position_children_centered(prop)

        except Exception as e:
            print(f"Error loading {title}: {e}")
            return self.workflow_generator.add_property(title=f"{title} (Error)")
    
    def _parse_node(self, parent_block, item):
        """Parse a single item from the vcl-plan structure - relies on actual JSON structure"""
        tag = item.get('tag', '')
        contents = item.get('contents', {})
        
        if tag == 'Disjunct':
            sub_items = contents.get('unDisjunctAll', [])
            or_block = self.workflow_generator.add_or(parent_block)
            for sub_item in sub_items:
                self._parse_node(or_block, sub_item)

        elif tag == 'Conjunct':
            sub_items = contents.get('unConjunctAll', [])
            and_block = self.workflow_generator.add_and(parent_block)
            for sub_item in sub_items:
                self._parse_node(and_block, sub_item)

        elif tag == 'Query':
            queries = contents.get('queries', {}).get('unDisjunctAll', [])
            negated = contents.get('negated', False)
            
            for _ in queries:
                self.global_query_id += 1
                query_path = os.path.join(CACHE_DIR, f"{self.property_name}-query{self.global_query_id}.txt")
                self.workflow_generator.add_query(self.global_query_id, parent_block, query_path, is_negated=negated)


class QueryTab(QTabWidget):
    """Verification node editor widget with property dropdown navigation"""
    property_changed = pyqtSignal(str)
    
    def __init__(self, parent=None, cache_location=CACHE_DIR):
        super().__init__(parent)
        self.cache_location = Path(cache_location)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        
        # Setup tab behavior
        self.setMovable(True)
        self.setTabsClosable(True)
        self.tabCloseRequested.connect(self._close_tab)
        self.tabBar().tabMoved.connect(self._refresh_close_buttons)
        self.currentChanged.connect(self._refresh_close_buttons)
        
        # Create workflow components
        self.graphics_scene = GraphicsScene()
        self.graphics_view = GraphicsView(self.graphics_scene)
        self.workflow_generator = VerificationWorkflow(self.graphics_scene, self.graphics_view)
        self.property_loader = PropertyLoader(self.workflow_generator)
        
        # Create and add workflow tab
        self._workflow_widget = self._create_workflow_widget()
        self.addTab(self._workflow_widget, "Workflow")
        self.setCurrentIndex(0)  # Ensure the workflow tab is selected
        self._refresh_close_buttons()
        
        # Connect signals and load initial property
        BlockGraphics.signals.query_double_clicked.connect(self._on_query_double_clicked)
        self._populate_property_dropdown()
        self._load_current_property()
    
    def _create_workflow_widget(self):
        """Create workflow widget with property dropdown + split editor pane."""
        widget = QWidget()
        layout = QVBoxLayout(widget)

        # Property selection header
        header = QHBoxLayout()
        header.addWidget(QLabel("Property:"))

        self.property_dropdown = QComboBox()
        self.property_dropdown.setMinimumWidth(150)
        self.property_dropdown.currentTextChanged.connect(self._on_property_changed)
        header.addWidget(self.property_dropdown)
        header.addStretch()

        layout.addLayout(header)

        # --- NEW: Splitter with left workflow view and right editor tabs ---
        self._splitter = QSplitter(Qt.Orientation.Horizontal)
        self._splitter.setChildrenCollapsible(False)

        # Left side: the existing graphics view (tree/workflow)
        left = QWidget()
        left_layout = QVBoxLayout(left)
        left_layout.setContentsMargins(0, 0, 0, 0)
        left_layout.addWidget(self.graphics_view)
        left.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Expanding)

        # Right side: editor tabs for opened query files
        self._editor_tabs = QTabWidget()
        self._editor_tabs.setTabsClosable(True)
        self._editor_tabs.tabCloseRequested.connect(self._close_editor_tab)
        self._editor_tabs.hide()  # hidden until first query opens
        self._editor_tabs.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Expanding)

        self._splitter.addWidget(left)
        self._splitter.addWidget(self._editor_tabs)

        self._splitter.setSizes([800, 0])

        layout.addWidget(self._splitter)
        layout.setContentsMargins(10, 10, 10, 10)
        layout.setSpacing(10)

        return widget
        
    def _populate_property_dropdown(self):
        """Populate dropdown with available properties"""
        properties = []
        self.property_dropdown.clear()

        if self.cache_location.exists():
            pattern = "*.vcl-plan"
            plan_files = list(self.cache_location.glob(pattern))
            properties.extend([f.stem for f in plan_files])

        self.property_dropdown.addItems(properties if properties else ["No properties found"])
    
    def _on_property_changed(self, property_name: str):
        """Handle property selection change"""
        if property_name and property_name != "No properties found":
            self._load_current_property()
            self.property_changed.emit(property_name)
    
    def _load_current_property(self):
        """Load the currently selected property"""
        property_name = self.property_dropdown.currentText()
        if property_name and property_name != "No properties found":
            self.property_loader.load_property(property_name, self.cache_location)
    
    def _on_query_double_clicked(self, file_path: str, tab_title: str):
        """Handle query double-click by opening file in new tab"""
        try:
            with open(file_path, 'r', encoding='utf-8') as file:
                content = file.read()
            self.add_text_tab(tab_title, content)
        except Exception as e:
            error_content = f"Error reading query file: {file_path}\n\nError: {str(e)}"
            self.add_text_tab(f"Error - {tab_title}", error_content)
    
    def add_text_tab(self, title: str, text: str) -> int:
        """Add a new text tab"""
        for i in range(self._editor_tabs.count()):
            if self._editor_tabs.tabText(i) == title:
                self._editor_tabs.setCurrentIndex(i)
                self._ensure_editor_visible()
                return i

        editor = QTextEdit()
        editor.setReadOnly(True)
        editor.setPlainText(text)
        idx = self._editor_tabs.addTab(editor, title)
        self._editor_tabs.setCurrentIndex(idx)
        self._ensure_editor_visible()
        return idx
    
    def clear(self):
        """Clear workflow and text tabs"""
        self.workflow_generator.clear_workflow()
        self._clear_text_tabs()
    
    def refresh_properties(self):
        """Refresh available properties list"""
        current = self.property_dropdown.currentText()
        self._populate_property_dropdown()
        
        # Restore selection if possible
        index = self.property_dropdown.findText(current)
        if index != -1:
            self.property_dropdown.setCurrentIndex(index)
        else:
            self.property_dropdown.setCurrentIndex(0)
        self._load_current_property()
    
    def _clear_text_tabs(self):
        """Remove all right-pane editor tabs."""
        for i in range(self._editor_tabs.count() - 1, -1, -1):
            w = self._editor_tabs.widget(i)
            self._editor_tabs.removeTab(i)
            if w:
                w.deleteLater()
        self._collapse_editor_if_needed()
    
    def _close_tab(self, index: int):
        """Close tab if not the workflow tab"""
        if 0 <= index < self.count():
            widget = self.widget(index)
            if widget is not self._workflow_widget:
                self.removeTab(index)
                widget.deleteLater()
    
    def _refresh_close_buttons(self):
        """Update close button visibility"""
        tb = self.tabBar()
        for i in range(self.count()):
            widget = self.widget(i)
            is_closable = widget is not self._workflow_widget   # Only workflow tab is not closable
            btn = tb.tabButton(i, QTabBar.ButtonPosition.RightSide)
            if btn:
                btn.setVisible(is_closable)
                btn.setEnabled(is_closable)

    def _ensure_editor_visible(self):
        """Show the right editor pane and give it space."""
        if self._editor_tabs.isHidden():
            self._editor_tabs.show()
        self._splitter.setSizes([400, 400])

    def _collapse_editor_if_needed(self):
        """Hide the right editor pane when no tabs remain."""
        if self._editor_tabs.count() == 0:
            self._editor_tabs.hide()
            # Collapse everything to the left again
            self._splitter.setSizes([800, 0])

    def _close_editor_tab(self, index: int):
        """Close a right-pane editor tab and collapse if none remain."""
        if 0 <= index < self._editor_tabs.count():
            w = self._editor_tabs.widget(index)
            self._editor_tabs.removeTab(index)
            if w:
                w.deleteLater()
            self._collapse_editor_if_needed()


"""
Module graphics_scene.py

This module contains the GraphicsScene class for handling graphics objects representation

Original author: Andrea Gimelli, Giacomo Rosato, Stefano Demarchi

"""

import math

from PyQt6.QtCore import QLine, pyqtSignal, QTimer, Qt
from PyQt6.QtGui import QColor, QPen
from PyQt6.QtWidgets import QGraphicsScene

import vehicle_gui.query_view.styling.dimension as dim
import vehicle_gui.query_view.styling.palette as palette


class GraphicsScene(QGraphicsScene):
    """
    This class manages the graphics scene, i.e., the handler of graphical objects. Objects in the
    GraphicsScene can be rendered in the viewport provided by the GraphicsView

    Attributes
    ----------

    Methods
    ----------

    """

    # The following signals are intercepted by QGraphicsScene.selectedItems()
    itemSelected = pyqtSignal()
    itemsDeselected = pyqtSignal()

    def __init__(self, scene=None, parent=None):
        super().__init__(parent)

        # Reference to the scene (optional)
        self.scene_ref = scene

        # Start with a smaller initial scene size
        self.min_scene_width = dim.INITIAL_SCENE_WIDTH
        self.min_scene_height = dim.INITIAL_SCENE_HEIGHT
        self.padding = dim.SCENE_PADDING 
        
        # Set initial smaller scene rectangle
        self.setSceneRect(-self.min_scene_width // 2, -self.min_scene_height // 2, 
                         self.min_scene_width, self.min_scene_height)

        self._color_background = QColor(palette.BACKGROUND_GREY)
        self._color_light = QColor(palette.BACKGROUND_LIGHT_LINE_GREY)
        self._color_dark = QColor(palette.BACKGROUND_DARK_LINE_GREY)

        # Pen settings
        self._pen_light = QPen(self._color_light)
        self._pen_light.setWidth(1)
        self._pen_dark = QPen(self._color_dark)
        self._pen_dark.setWidth(2)

        self.setBackgroundBrush(self._color_background)
        self._size_adjustment_pending = False  # Flag to track size adjustment state

    def auto_adjust_scene_size(self):
        """Automatically adjust scene size to fit all items with padding"""
        if not self.items():
            # No items, keep minimum size
            self.setSceneRect(-self.min_scene_width // 2, -self.min_scene_height // 2, 
                             self.min_scene_width, self.min_scene_height)
            return

        # Calculate the bounding rect of all items
        items_rect = self.itemsBoundingRect()
        
        # Add padding around the content
        content_left = items_rect.left() - self.padding
        content_top = items_rect.top() - self.padding
        content_right = items_rect.right() + self.padding
        content_bottom = items_rect.bottom() + self.padding
        
        # Calculate required dimensions
        required_width = content_right - content_left
        required_height = content_bottom - content_top
        
        # Ensure minimum size
        scene_width = max(required_width, self.min_scene_width)
        scene_height = max(required_height, self.min_scene_height)
        
        # Center the scene around the content
        center_x = (content_left + content_right) / 2
        center_y = (content_top + content_bottom) / 2
        
        # Set new scene rect centered on content
        new_left = center_x - scene_width / 2
        new_top = center_y - scene_height / 2
        
        self.setSceneRect(new_left, new_top, scene_width, scene_height)

    def addItem(self, item):
        """Override addItem to auto-adjust scene size"""
        super().addItem(item)
        # Defer the size adjustment to avoid performance issues during bulk additions
        if not self._size_adjustment_pending:
            self._size_adjustment_pending = True
            QTimer.singleShot(0, self._perform_size_adjustment)     # Batch size adjustments

    def removeItem(self, item):
        """Override removeItem to auto-adjust scene size"""
        super().removeItem(item)
        # Defer the size adjustment
        if not self._size_adjustment_pending:
            self._size_adjustment_pending = True
            QTimer.singleShot(0, self._perform_size_adjustment)

    def _perform_size_adjustment(self):
        """Perform the actual size adjustment (called via QTimer)"""
        if self._size_adjustment_pending:
            self._size_adjustment_pending = False
        self.auto_adjust_scene_size()

    def clear(self):
        """Override clear to reset scene size"""
        super().clear()
        # Reset to minimum size when clearing
        self.setSceneRect(-self.min_scene_width // 2, -self.min_scene_height // 2, 
                         self.min_scene_width, self.min_scene_height)

    def dragMoveEvent(self, event):
        """
        Necessary override for enabling events

        """

        pass

    # Event handlers to ignore right-click events and allow propagation to view

    def mousePressEvent(self, event):
        """Handle mouse press events in the scene"""
        # Don't consume right mouse events
        if event.button() == Qt.MouseButton.RightButton:
            event.ignore() 
        else:
            super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        """Handle mouse move events in the scene"""
        # Don't consume right mouse events during dragging
        if event.buttons() & Qt.MouseButton.RightButton:
            event.ignore()
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        """Handle mouse release events in the scene"""
        # Don't consume right mouse events
        if event.button() == Qt.MouseButton.RightButton:
            event.ignore()
        else:
            super().mouseReleaseEvent(event)

    def drawBackground(self, painter, rect):
        """
        This method draws the background of the scene (setting the color and adding a grid)
        using the painter and a set of QPens

        Parameters
        ----------
        painter : QPainter
            QPainter performs low-level painting on widgets and other paint devices
        rect : QRectF
            A rectangle is normally expressed as a top-left corner and a size

        """

        super().drawBackground(painter, rect)

        # Here we create our grid
        left = int(math.floor(rect.left()))
        right = int(math.ceil(rect.right()))
        top = int(math.floor(rect.top()))
        bottom = int(math.ceil(rect.bottom()))

        first_left = left - (left % dim.GRID_SIZE)
        first_top = top - (top % dim.GRID_SIZE)

        # Compute all lines to be drawn
        lines_light, lines_dark = [], []  # position (x1, y1), (x2, y2)

        for x in range(first_left, right, dim.GRID_SIZE):
            if x % (dim.GRID_SIZE * dim.GRID_SQUARE) != 0:
                lines_light.append(QLine(x, top, x, bottom))
            else:
                lines_dark.append(QLine(x, top, x, bottom))

        for y in range(first_top, bottom, dim.GRID_SIZE):
            if y % (dim.GRID_SIZE * dim.GRID_SQUARE) != 0:
                lines_light.append(QLine(left, y, right, y))
            else:
                lines_dark.append(QLine(left, y, right, y))

        painter.setPen(self._pen_light)
        painter.drawLines(*lines_light)
        painter.setPen(self._pen_dark)
        painter.drawLines(*lines_dark)



"""
Module graphics_view.py

This module contains the GraphicsView class for rendering graphics objects in the viewport

Original author: Andrea Gimelli, Giacomo Rosato, Stefano Demarchi

"""

from PyQt6.QtCore import Qt, QTimer, QPointF, QRectF, QPropertyAnimation, QObject
from PyQt6.QtGui import QPainter
from PyQt6 import QtGui
from PyQt6.QtWidgets import QGraphicsView

import vehicle_gui.query_view.styling.dimension as dim
from .graphics_scene import GraphicsScene


class GraphicsView(QGraphicsView):
    """
    This class visualizes the contents of the GraphicsScene in a scrollable viewport

    """

    def __init__(self, gr_scene: 'GraphicsScene', parent=None):
        super().__init__(parent)

        # Reference to the graphics scene
        self.gr_scene_ref = gr_scene
        self.setScene(self.gr_scene_ref)
        self.zoom = dim.ZOOM

        # Right mouse drag state
        self.right_mouse_dragging = False
        self.last_mouse_pos = None

        self.init_ui()

    def init_ui(self):
        self.setViewportUpdateMode(QGraphicsView.ViewportUpdateMode.FullViewportUpdate)

        self.setRenderHints(QPainter.RenderHint.Antialiasing | QPainter.RenderHint.TextAntialiasing |
                            QPainter.RenderHint.SmoothPixmapTransform)
        self.setDragMode(QGraphicsView.DragMode.RubberBandDrag)
        self.setTransformationAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)

        # Enable scrollbars
        self.h_scrollbar = self.horizontalScrollBar()
        self.v_scrollbar = self.verticalScrollBar()
        
        # Ensure the view can receive mouse events
        self.setMouseTracking(True)
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.setAttribute(Qt.WidgetAttribute.WA_AcceptTouchEvents, False)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.NoContextMenu)  # Disable right-click context menu

    def zoom_in(self):
        self.zoom += dim.ZOOM_STEP
        self.set_scale(dim.ZOOM_IN_FACTOR)

    def zoom_out(self):
        self.zoom -= dim.ZOOM_STEP
        self.set_scale(1 / dim.ZOOM_IN_FACTOR)

    def set_scale(self, factor: float):
        clipped = False
        if self.zoom < dim.ZOOM_RANGE[0]:
            self.zoom = dim.ZOOM_RANGE[0]
            clipped = True
        if self.zoom > dim.ZOOM_RANGE[1]:
            self.zoom = dim.ZOOM_RANGE[1]
            clipped = True

        # Set scene scale
        if not clipped:
            self.scale(factor, factor)

    def delete_items(self, sel_ids: list):
        for block_id in sel_ids:
            block = self.gr_scene_ref.scene_ref.blocks[block_id]
            self.gr_scene_ref.scene_ref.remove_block(block, logic=True)

    def mousePressEvent(self, event: 'QtGui.QMouseEvent') -> None:
        """
        Handle mouse press events, including right mouse button for dragging
        """

        if event.button() == Qt.MouseButton.RightButton:
            self.setFocus()
            self.right_mouse_dragging = True
            self.last_mouse_pos = event.position()
            self.setCursor(Qt.CursorShape.ClosedHandCursor)
            event.accept()
        else:
            super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: 'QtGui.QMouseEvent') -> None:
        """
        Handle mouse release events
        """

        if event.button() == Qt.MouseButton.RightButton:
            self.right_mouse_dragging = False
            self.last_mouse_pos = None
            self.setCursor(Qt.CursorShape.ArrowCursor)
            event.accept()
        else:
            super().mouseReleaseEvent(event)

    def mouseMoveEvent(self, event: 'QtGui.QMouseEvent') -> None:
        """
        Handle mouse move events, including right mouse dragging
        """

        if self.right_mouse_dragging and self.last_mouse_pos is not None:
            # Calculate the delta movement
            delta = event.position() - self.last_mouse_pos
            
            # Scroll the view by the delta amount
            self.horizontalScrollBar().setValue(int(self.horizontalScrollBar().value() - delta.x()))
            self.verticalScrollBar().setValue(int(self.verticalScrollBar().value() - delta.y()))
            
            # Update the last position
            self.last_mouse_pos = event.position()
            event.accept()
        else:
            super().mouseMoveEvent(event)

    def wheelEvent(self, event: 'QtGui.QWheelEvent') -> None:
        """
        Override the event to enable zoom

        """

        factor = 1 / dim.ZOOM_IN_FACTOR

        if event.angleDelta().y() > 0:
            factor = dim.ZOOM_IN_FACTOR
            self.zoom += dim.ZOOM_STEP
        else:
            self.zoom -= dim.ZOOM_STEP

        self.set_scale(factor)

    def scroll_item_into_view(self, item, margin_scene=20.0):
        """
        Ensure `item` is fully visible in `view`.
        - Pans horizontally if the item is out of view horizontally.
        - Pans vertically if the item is out of view vertically.
        - `margin_scene` is in scene units (zoom-independent).
        - If `animate=True`, does a short smooth pan (assumes scale-only transform).
        """
        def _do():
            # Check if item is still valid before accessing it
            try:
                if not item or not hasattr(item, 'mapToScene') or not hasattr(item, 'boundingRect'):
                    return
                
                # Item bounds in scene coords
                item_rect: QRectF = item.mapToScene(item.boundingRect()).boundingRect()
                view_rect: QRectF = self.mapToScene(self.viewport().rect()).boundingRect()

                dx = 0.0
                dy = 0.0

                # Horizontal adjustment
                if item_rect.left() - margin_scene < view_rect.left():
                    dx = (item_rect.left() - margin_scene) - view_rect.left()
                elif item_rect.right() + margin_scene > view_rect.right():
                    dx = (item_rect.right() + margin_scene) - view_rect.right()

                # Vertical adjustment
                if item_rect.top() - margin_scene < view_rect.top():
                    dy = (item_rect.top() - margin_scene) - view_rect.top()
                elif item_rect.bottom() + margin_scene > view_rect.bottom():
                    dy = (item_rect.bottom() + margin_scene) - view_rect.bottom()
            # Ignore if item has been deleted
            except (RuntimeError, AttributeError) as e:
                return

            if dx == 0.0 and dy == 0.0:
                return

            # Compute target viewport rect
            new_left   = view_rect.left() + dx
            new_top    = view_rect.top() + dy
            new_width  = view_rect.width()
            new_height = view_rect.height()

            # Clamp to scene
            scene_rect = self.sceneRect()
            if scene_rect.isValid():
                new_left = max(scene_rect.left(), min(new_left, scene_rect.right() - new_width))
                new_top  = max(scene_rect.top(),  min(new_top,  scene_rect.bottom() - new_height))

            # Convert to a target center and pan
            target_center = QPointF(new_left + new_width / 2.0, new_top + new_height / 2.0)
            self.centerOn(target_center)

        QTimer.singleShot(0, _do)



from PyQt6.QtWidgets import QPlainTextEdit, QWidget, QToolTip
from PyQt6.QtCore import Qt, QRect, QSize
from PyQt6.QtGui import QColor, QPainter, QFontDatabase
from PyQt6.QtGui import QPen, QTextCharFormat
from superqt.utils import CodeSyntaxHighlight
from collections import defaultdict


class ExtendedSyntaxHighlight(CodeSyntaxHighlight):
    """Syntax highlighter for Vehicle language with error highlighting"""
    def __init__(self, parent, lang, theme):
        self.line_errors = defaultdict(list)
        super().__init__(parent, lang, theme)

        # Squiggle format for errors
        self.error_format = QTextCharFormat()
        self.error_format.setUnderlineColor(QColor("red"))
        self.error_format.setUnderlineStyle(QTextCharFormat.UnderlineStyle.SpellCheckUnderline)

    def set_errors(self, errors: list[dict]):
        """Set the list of errors to highlight"""
        self.line_errors.clear()
        for err in  errors:
            line = err["provenance"]["contents"][0]
            self.line_errors[line].append(err)
        self.rehighlight()

    def highlightBlock(self, text):
        """Highlight the current block, adding error highlights if needed"""
        super().highlightBlock(text)
        if not text: return

        block_number = self.currentBlock().blockNumber() + 1

        # Reset formatting
        current_format = self.format(0)
        combined_format = QTextCharFormat(current_format)
        combined_format.setUnderlineStyle(QTextCharFormat.UnderlineStyle.NoUnderline)
        self.setFormat(0, len(text), combined_format)

        for err in self.line_errors.get(block_number, []):
            start_line, start_col, end_line, end_col = err["provenance"]["contents"]
            if block_number == start_line and block_number == end_line:     # Single-line error
                start_idx = max(0, start_col - 1)
                end_idx = min(len(text) - 1, end_col - 1)
            elif block_number == start_line:                                # First line of multi-line error
                start_idx = max(0, start_col - 1)
                end_idx = len(text) - 1
            elif block_number == end_line:                                  # Last line of multi-line error
                start_idx = 0
                end_idx = min(len(text) - 1, end_col - 1)
            elif start_line < block_number < end_line:                      # Middle line of multi-line error
                start_idx = 0
                end_idx = len(text) - 1
            else:
                continue                                                    # No error on this line
            
            # Apply error formatting
            length = max(0, end_idx - start_idx + 1)
            combined_format.merge(self.error_format)
            self.setFormat(start_idx, length, combined_format)

            
class CodeEditor(QPlainTextEdit):
    """Custom editor with line numbers"""
    def __init__(self, lang, theme, parent=None):
        super().__init__(parent)

        # Use a fixed-width font everywhere
        mono = QFontDatabase.systemFont(QFontDatabase.FixedFont)
        mono.setPointSize(14)
        self.setFont(mono)

        # Create inline line-number area
        self.line_number_area = QWidget(self)

        # Bind the size hint and paint events
        self.line_number_area.sizeHint = lambda: QSize(self.line_number_area_width(), 0)
        self.line_number_area.paintEvent = self.line_number_area_paint_event

        # Connect signals
        self.blockCountChanged.connect(self.update_line_number_area_width)
        self.updateRequest.connect(self.update_line_number_area)
        self.cursorPositionChanged.connect(self.highlight_current_line)

        # Initial update
        self.update_line_number_area_width(0)
        self.highlight_current_line()

        # Syntax highlighting
        self.highlighter = ExtendedSyntaxHighlight(self.document(), lang, theme)
        self.setMouseTracking(True)
        self.viewport().installEventFilter(self)

    def eventFilter(self, obj, event):
        """Show tooltips for errors on hover"""
        if event.type() == event.Type.MouseMove:
            cursor = self.cursorForPosition(event.position().toPoint())
            block_number = cursor.blockNumber() + 1
            col = cursor.positionInBlock() + 1

            tooltip_shown = False
            messages = []

            errors = self.highlighter.line_errors.get(block_number, [])
            for err in errors:
                _, start_col, _, end_col = err["provenance"]["contents"]
                if start_col <= col <= end_col:
                    tooltip_shown = True
                    messages.append(f"Problem: {err['problem']}\nFix: {err['fix']}")
                    break

            if tooltip_shown:
                QToolTip.showText(event.globalPosition().toPoint(), "\n".join(messages))
            else:
                QToolTip.hideText() 
        
        return super().eventFilter(obj, event)

    def add_errors(self, errors: list[dict]):
        self.highlighter.set_errors(errors)

    def clear_errors(self):
        self.highlighter.set_errors([])

    def line_number_area_width(self):
        """Calculate the width of the line number area"""
        digits = 1
        max_num = max(1, self.blockCount())
        while max_num >= 10:
            max_num //= 10
            digits += 1
        space = self.fontMetrics().horizontalAdvance('9') * digits
        padding = 4
        return space + 2 * padding

    def update_line_number_area_width(self, new_block_count):
        """Update the margin according to the line number area width"""
        self.setViewportMargins(self.line_number_area_width(), 0, 0, 0)

    def update_line_number_area(self, rect, dy):
        """Update the line number area when needed"""
        if dy:
            self.line_number_area.scroll(0, dy)
        else:
            self.line_number_area.update(0, rect.y(), self.line_number_area.width(), rect.height())
        if rect.contains(self.viewport().rect()):
            self.update_line_number_area_width(0)

    def resizeEvent(self, event):
        """Handle resize event"""
        super().resizeEvent(event)
        cr = self.contentsRect()
        self.line_number_area.setGeometry(QRect(cr.left(), cr.top(), self.line_number_area_width(), cr.height()))

    def line_number_area_paint_event(self, event):
        """Paint the line numbers"""
        painter = QPainter(self.line_number_area)
        # Draw line numbers in the same monospaced font
        painter.setFont(self.font())
        block = self.firstVisibleBlock()
        block_number = block.blockNumber()
        top = self.blockBoundingGeometry(block).translated(self.contentOffset()).top()
        bottom = top + self.blockBoundingRect(block).height()
        while block.isValid() and top <= event.rect().bottom():
            if block.isVisible() and bottom >= event.rect().top():
                number = str(block_number + 1)
                painter.setPen(QColor(150, 150, 150))
                padding = 4
                painter.drawText(
                    padding,
                    int(top),
                    self.line_number_area.width() - 2 * padding,
                    self.fontMetrics().height(),
                    Qt.AlignmentFlag.AlignRight,
                    number
                )
            block = block.next()
            top = bottom
            bottom = top + self.blockBoundingRect(block).height()
            block_number += 1

    def highlight_current_line(self):
        """Trigger repaint to draw horizontal rules around the current line"""
        self.viewport().update()

    def paintEvent(self, event):
        """Paint editor normally, then draw rules above and below the current line"""
        super().paintEvent(event)
        painter = QPainter(self.viewport())

        color = QColor(200, 200, 200, 20)  
        pen = QPen(color, 2, Qt.SolidLine)
        painter.setPen(pen)

        cursor = self.textCursor()
        block = cursor.block()
        block_geo = self.blockBoundingGeometry(block).translated(self.contentOffset())
        top = block_geo.top()
        bottom = top + self.blockBoundingRect(block).height() - 4
        
        painter.drawLine(0, int(top), self.viewport().width(), int(top))
        painter.drawLine(0, int(bottom), self.viewport().width(), int(bottom))


from pygments.style import Style
from pygments.token import Token, Comment, Keyword, Name, String, \
     Error, Generic, Number, Operator

class VSE(Style):
    styles = {
        Token:                  '',
        Comment:                '#3a7f1a',  # Green
        Keyword:                '#2257e7',  # Blue
        Name:                   '#12ba87',  # Light Green
        String:                 '#df6400',
        Number:                 '#b2f883',  # Pale Green
        Operator:               '#ffffff'   # White
    }



"""
Verification package for the node editor.

This package contains all verification-specific logic including:
- Verification scene and workflow models
- Verification workflow generators
- Verification-specific block types

"""

from .workflow import VerificationWorkflow
from .blocks import PropertyBlock, QueryBlock, WitnessBlock

__all__ = [
    'VerificationWorkflow',
    'PropertyBlock',
    'QueryBlock',
    'WitnessBlock'
]



"""
Module block_graphics.py

This module contains the graphics elements of Block objects for representing the high-level properties,
low-level queries, and witnesses in the verification output

This module was migrated from the network node editor in CoCoNet.

Original author: Andrea Gimelli, Giacomo Rosato, Stefano Demarchi

"""

from PyQt6 import QtCore
from PyQt6.QtCore import Qt, pyqtSignal, QObject
from PyQt6.QtGui import QPen, QColor, QFont, QBrush, QPainterPath, QPainter
from PyQt6.QtWidgets import QGraphicsItem, QWidget, QGraphicsProxyWidget, QGraphicsTextItem, QVBoxLayout, QGridLayout, \
    QHBoxLayout, QMessageBox, QStyleOptionGraphicsItem, QGraphicsSceneHoverEvent, QGraphicsSceneMouseEvent

import vehicle_gui.query_view.styling.dimension as dim
import vehicle_gui.query_view.styling.palette as palette
from ..verification.blocks import QueryBlock
from ..base_types import Block


class BlockSignals(QObject):
    """Signal emitter for block events"""
    query_double_clicked = pyqtSignal(str, str)  # path, title


class BlockGraphics(QGraphicsItem):
    """Graphics representation of a Block domain model"""
    signals = BlockSignals()     
    
    def __init__(self, block: 'Block', parent=None):
        super().__init__(parent)
        # Reference to the block domain model
        self.block_ref = block

        # Hover flag
        self.hover = False

        # Content widget
        self.content = None
        self.width = dim.BLOCK_BASE_WIDTH
        self.height = dim.BLOCK_BASE_HEIGHT

        # Init graphics content
        self.title_item = QGraphicsTextItem(self)
        self.graphics_content = QGraphicsProxyWidget(self)
        self.init_graphics_content()

        # Style parameters
        self.color_scheme = []
        self.init_colors()
        self._pen_default = QPen(QColor(self.color_scheme[0]))
        self._pen_default.setWidth(2)
        self._pen_hovered = QPen(QColor(self.color_scheme[1]))
        self._pen_hovered.setWidth(2)
        self._pen_selected = QPen(QColor(self.color_scheme[2]))
        self._pen_selected.setWidth(3)
        self._pen_selected.setStyle(Qt.PenStyle.DotLine)
        self._brush_title = QBrush(QColor(self.color_scheme[3]))
        self._brush_background = QBrush(QColor(self.color_scheme[4]))

        self.init_flags()

    def init_title(self):
        """
        This method sets up the title widget

        """

        self.title_item.setDefaultTextColor(QColor(palette.WHITE))
        self.title_item.setFont(QFont(dim.FONT_FAMILY, dim.FONT_SIZE))
        self.title_item.setPos(dim.TITLE_PAD, 0)
        self.title_item.setPlainText(self.block_ref.title)
        self.title_item.setTextWidth(self.width - 2 * dim.TITLE_PAD)

    def init_colors(self):
        """
        This method sets up the color scheme of the block
        using polymorphic behavior from the block itself
        """
        self.color_scheme = self.block_ref.get_color_scheme()

    def init_graphics_content(self):
        """
        This method sets up the graphics properties of the block
        depending on the content
        """
        if self.block_ref.has_parameters():
            self.width = dim.BLOCK_PARAM_WIDTH
        else:
            self.width = self.block_ref.get_block_width()

        # Init title card after setting the correct widget width
        self.init_title()
        if isinstance(self.block_ref, QueryBlock):
            self.setToolTip("Double click to view the query")

    def init_flags(self):
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)
        self.setAcceptHoverEvents(True)

    def open_dock_params(self):
        self.block_ref.scene_ref.editor_widget_ref.show_inspector()

    def hoverEnterEvent(self, event: 'QGraphicsSceneHoverEvent') -> None:
        self.hover = True
        if isinstance(self.block_ref, QueryBlock):
            self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.update()

    def hoverLeaveEvent(self, event: 'QGraphicsSceneHoverEvent') -> None:
        self.hover = False
        self.setCursor(Qt.CursorShape.ArrowCursor)
        self.update()

    def mouseMoveEvent(self, event: 'QGraphicsSceneMouseEvent') -> None:
        super().mouseMoveEvent(event)

        for block in self.block_ref.scene_ref.blocks.values():
            if block.graphics.isSelected():
                block.update_edges()

        self.block_ref.update_edges()

    def mouseDoubleClickEvent(self, event: 'QGraphicsSceneMouseEvent') -> None:
        if isinstance(self.block_ref, QueryBlock) and hasattr(self.block_ref, "path"):
            # Emit signal to hierarchical output
            tab_title = f"{self.block_ref.title} - {self.block_ref.path.split('/')[-1]}"
            self.signals.query_double_clicked.emit(self.block_ref.path, tab_title)


    def paint(self, painter: 'QPainter', option: 'QStyleOptionGraphicsItem', widget=None) -> None:
        """
        This method draws the graphicsBlock item. It is a rounded rectangle divided in 3 sections:

        Outline section: draw the contours of the block
        Title section: a darker rectangle in which lays the title
        Content section: container for the block parameters

        """

        # Title section
        path_title = QPainterPath()
        path_title.setFillRule(Qt.FillRule.WindingFill)
        path_title.addRoundedRect(0, 0, self.width, dim.TITLE_HEIGHT, dim.EDGE_ROUNDNESS, dim.EDGE_ROUNDNESS)

        # Remove bottom rounded corners for title box
        path_title.addRect(0, dim.TITLE_HEIGHT - dim.EDGE_ROUNDNESS, dim.EDGE_ROUNDNESS, dim.EDGE_ROUNDNESS)
        path_title.addRect(self.width - dim.EDGE_ROUNDNESS, dim.TITLE_HEIGHT - dim.EDGE_ROUNDNESS,
                           dim.EDGE_ROUNDNESS, dim.EDGE_ROUNDNESS)
        painter.setPen(Qt.PenStyle.NoPen)
        painter.setBrush(self._brush_title)
        painter.drawPath(path_title.simplified())

        # Content
        path_content = QPainterPath()
        path_content.setFillRule(Qt.FillRule.WindingFill)
        path_content.addRoundedRect(0, dim.TITLE_HEIGHT, self.width,
                                    self.height - dim.TITLE_HEIGHT, dim.EDGE_ROUNDNESS, dim.EDGE_ROUNDNESS)

        # Remove top rounded corners for content box
        # left
        path_content.addRect(0, dim.TITLE_HEIGHT, dim.EDGE_ROUNDNESS, dim.EDGE_ROUNDNESS / 2)
        # right
        path_content.addRect(self.width - dim.EDGE_ROUNDNESS, dim.TITLE_HEIGHT,
                             dim.EDGE_ROUNDNESS, dim.EDGE_ROUNDNESS / 2)
        painter.setPen(Qt.PenStyle.NoPen)
        painter.setBrush(self._brush_background)
        painter.drawPath(path_content.simplified())

        # Outline
        path_outline = QPainterPath()
        path_outline.addRoundedRect(0, 0, self.width, self.height, dim.EDGE_ROUNDNESS, dim.EDGE_ROUNDNESS)
        painter.setPen(self._pen_default if not self.isSelected() else self._pen_selected)

        if self.hover and not self.isSelected():
            painter.setPen(self._pen_hovered)

        painter.setBrush(Qt.BrushStyle.NoBrush)
        painter.drawPath(path_outline.simplified())

    def boundingRect(self) -> QtCore.QRectF:
        """
        Defines the Qt bounding rectangle

        Returns
        ----------
        QRectF
            The area in which the click triggers the item

        """

        return QtCore.QRectF(0, 0, self.width, self.height).normalized()



import os
import site
import traceback 
import asyncio
from PyQt6.QtWidgets import (QMainWindow, QTextEdit, QVBoxLayout, QPushButton, QWidget,
                             QLabel, QFileDialog, QHBoxLayout, QStatusBar, QMessageBox,
                             QScrollArea, QSizePolicy, QToolBar, QFrame, QSplitter,
                             QTabWidget, QProgressBar, QApplication, QComboBox)
from PyQt6.QtCore import Qt, QRunnable, pyqtSlot, QObject, pyqtSignal, QThreadPool, QTimer
from PyQt6.QtGui import QFontDatabase, QIcon
from superqt.utils import CodeSyntaxHighlight
import functools
from typing import Callable
from pathlib import Path
import json
import math

from vehicle_gui.code_editor import CodeEditor
from vehicle_gui.vcl_bindings import VCLBindings
from vehicle_gui.query_view.query_tab import QueryTab
from vehicle_gui.counter_example_view.counter_example_tab import CounterExampleTab
from vehicle_gui.resource_view.input_view import InputView
from vehicle_gui.resource_view.property_view import PropertyView
from vehicle_gui.vcl_bindings import CACHE_DIR
from vehicle_gui.counter_example_view.base_renderer import GSImageRenderer, TextRenderer
from vehicle_gui.util import which_all

from vehicle_lang import VERSION 

RELEASE_VERSION = "0.3.1"

class OperationSignals(QObject):
    """
    Defines signals to communicate from worker thread to main GUI thread.
    """
    output_chunk = pyqtSignal(str, str)  # tag ('stdout'/'stderr'), chunk_text
    finished = pyqtSignal(int)           # return_code (0=success, 1=error, -1=stopped)
    error = pyqtSignal(str)              # For critical errors in worker setup itself


class OperationWorker(QRunnable):
    def __init__(self, operation: Callable, vcl_bindings: VCLBindings,
                 stop_event: asyncio.Event, signals: OperationSignals):
        super().__init__()
        self.operation = operation
        self.vcl_bindings = vcl_bindings
        self.stop_event = stop_event
        self.signals = signals

    def _callback_fn(self, tag: str, chunk: str):
        self.signals.output_chunk.emit(tag, chunk)

    def _finish_fn(self, return_code: int):
        if self.stop_event.is_set():
            self.signals.finished.emit(-1)  # -1 for user-stopped
        else:
            self.signals.finished.emit(return_code)

    @pyqtSlot()
    def run(self):
        try:
            self.operation(
                callback_fn=self._callback_fn,
                finish_fn=self._finish_fn,
                stop_event=self.stop_event
            )
        except Exception as e:
            tb_str = traceback.format_exc()
            self.signals.output_chunk.emit("stderr", f"Critical Worker Error: {e}\n{tb_str}")
            if self.stop_event.is_set(): # If stop was also requested
                self.signals.finished.emit(-1)
            else:
                self.signals.finished.emit(1)

    
class VehicleGUI(QMainWindow):
    """Vehicle GUI"""
    def __init__(self):
        super().__init__()
        self.stop_event = asyncio.Event()
        self.vcl_bindings = VCLBindings()
        self.vcl_path = None
        self.verifier_paths = {}  # Store mapping of verifier display names to paths
        self.setWindowTitle("Vehicle GUI")
        self.setGeometry(100, 100, 1400, 800)
        self.current_operation = None  # Tracks 'compile' or 'verify'
        # Buffer for incoming JSON chunks and per-property progress tracking
        self._json_buffer = ""
        self.current_property_name = ""
        self.property_total_queries = 0
        self.property_completed_queries = 0
        # Pause duration before starting next property (ms)
        self.property_pause_ms = 300
        self._waiting_transition = False

        self.thread_pool = QThreadPool()
        self.operation_signals = OperationSignals()
        self.operation_signals.output_chunk.connect(self._gui_process_output_chunk)
        self.operation_signals.finished.connect(self._gui_operation_finished)

        self.show_ui()
        # Verifier dropdown will automatically populate and set default verifier

    def show_ui(self):
        """Initialize UI"""
        # Create menu bar
        file_toolbar = QToolBar("File")
        self.addToolBar(file_toolbar)
        file_toolbar.setMovable(False)
        file_toolbar.setFloatable(False)
        file_toolbar.addAction(QIcon.fromTheme("document-new"), "New", self.new_file)
        file_toolbar.addAction(QIcon.fromTheme("document-open"), "Open", self.open_file)
        file_toolbar.addAction(QIcon.fromTheme("document-save"), "Save", self.save_file)

        # Add a spacer to the toolbar. This will push the buttons to the right
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        file_toolbar.addWidget(spacer)

        # Add compile, and verify buttons
        self.compile_button = QPushButton(QIcon.fromTheme("scanner"), "Compile")
        self.compile_button.clicked.connect(self.compile_spec)
        file_toolbar.addWidget(self.compile_button)

        self.verify_button = QPushButton(QIcon.fromTheme("media-playback-start"), "Verify")
        self.verify_button.clicked.connect(self.verify_spec)
        self.verify_button.setEnabled(False)
        file_toolbar.addWidget(self.verify_button)

        # Add stop button for running operations
        self.stop_button = QPushButton(QIcon.fromTheme("process-stop", QIcon.fromTheme("media-playback-stop")), "Stop")
        self.stop_button.setToolTip("Stop the current operation")
        self.stop_button.setEnabled(False)
        self.stop_button.clicked.connect(self.stop_current_operation)
        file_toolbar.addWidget(self.stop_button)

        # Create main window widget 
        main_widget = QSplitter(Qt.Orientation.Vertical)
        # Create top tabbed section
        main_tab = QTabWidget()
        main_widget.addWidget(main_tab)
        # Create each of the tabs
        input_tab = QWidget()
        output_tab = QWidget()
        main_tab.addTab(input_tab, "Specification")
        main_tab.addTab(output_tab, "Queries")

        # Define the splitter as the main widget
        self.setCentralWidget(main_widget)
        
        # Define layouts for each tab
        input_layout = QVBoxLayout(input_tab)
        output_layout = QVBoxLayout(output_tab)

        # Create main edit area 
        input_edit_layout = QHBoxLayout()

        # Create left area, containing the editor and the console
        left_layout = QVBoxLayout()
        left_label = QLabel("Editor")
        left_label.setAlignment(Qt.AlignmentFlag.AlignLeft)
        font = left_label.font()
        font.setPointSize(14)
        left_label.setFont(font)
        left_layout.addWidget(left_label)

        # Create a splitter for the editor and the console
        editor_console_splitter = QSplitter(Qt.Orientation.Vertical)

        # Create left editor 
        self.editor = CodeEditor(lang="external", theme="vse-style")
        mono = QFontDatabase.systemFont(QFontDatabase.FixedFont)
        mono.setPointSize(14)
        self.editor.setFont(mono)
        self.editor.setPlaceholderText("Enter your Vehicle specification here...")
        editor_console_splitter.addWidget(self.editor) # Add editor to splitter

        # Create the new console area, containing the problems and output tabs
        self.console_tab_widget = QTabWidget()
        console_font = QFontDatabase.systemFont(QFontDatabase.FixedFont)
        console_font.setPointSize(12)

        # Create the Problems tab in the console
        self.problems_console = QTextEdit()
        self.problems_console.setFont(console_font)
        self.problems_console.setReadOnly(True)
        self.console_tab_widget.addTab(self.problems_console, "Problems")

        # Create the Output tab in the console
        self.log_console = QTextEdit() # For "Output" tab
        self.log_console.setFont(console_font)
        self.log_console.setReadOnly(True)
        self.console_tab_widget.addTab(self.log_console, "Output")

        # Add console to splitter
        main_widget.addWidget(self.console_tab_widget)

        # Create the Counter Examples tab
        self.counter_example_tab = CounterExampleTab()
        main_tab.addTab(self.counter_example_tab, " Results")

        # Set the size policy for the editor and the console: editor takes 3/4 of the space
        editor_console_splitter.setStretchFactor(0, 3)
        editor_console_splitter.setStretchFactor(1, 1)

        # Add the splitter
        left_layout.addWidget(editor_console_splitter) 

        # Add the left layout to the main layout
        input_edit_layout.addLayout(left_layout, 3)

        # Create right area for resource boxes and output
        right_layout = QVBoxLayout()
        right_label = QLabel("Input")
        right_label.setAlignment(Qt.AlignmentFlag.AlignLeft)
        font = right_label.font()
        font.setPointSize(14)
        right_label.setFont(font)
        right_layout.addWidget(right_label)

        # Create resource view widget
        self.input_view = InputView(error_callback=self.append_to_problems)
        right_layout.addWidget(self.input_view)

        # Property selection widget
        properties_label = QLabel("Properties and Variables")
        properties_label.setAlignment(Qt.AlignmentFlag.AlignLeft)
        font = properties_label.font()
        font.setPointSize(14)
        properties_label.setFont(font)
        right_layout.addWidget(properties_label)

        self.property_view = PropertyView()
        self.property_view.renderers_changed.connect(self.update_counter_example_modes)
        right_layout.addWidget(self.property_view)
        
        # Create output box
        self.query_tab = QueryTab()
        output_layout.addWidget(self.query_tab)

        # Set size policy for output box
        self.input_view.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.query_tab.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.input_view.setMinimumWidth(200)
        input_edit_layout.addLayout(right_layout, 2)
        input_layout.addLayout(input_edit_layout)

        # Create status bar
        self.status_bar = QStatusBar()
        font = self.status_bar.font()
        font.setPointSize(12)
        self.status_bar.setFont(font)
        self.status_bar.setSizeGripEnabled(False)
        self.status_bar.setContentsMargins(0, 0, 0, 2)
        self.setStatusBar(self.status_bar)

        # File path label
        self.file_path_label = QLabel("No File Open")
        self.file_path_label.setContentsMargins(8, 0, 0, 0)
        self.file_path_label.setCursor(Qt.CursorShape.PointingHandCursor)
        self.file_path_label.mouseReleaseEvent = lambda event: self.open_file()
        self.status_bar.addWidget(self.file_path_label)

        # Separator between file display and cursor position
        sep_fd_cursor = QFrame()
        sep_fd_cursor.setFrameShape(QFrame.Shape.VLine)
        sep_fd_cursor.setFrameShadow(QFrame.Shadow.Sunken)
        self.status_bar.addWidget(sep_fd_cursor)

        # Cursor position label
        self.position_label = QLabel("Ln 1, Col 1")
        self.position_label.setContentsMargins(5, 0, 5, 0)
        self.status_bar.addWidget(self.position_label)

        # Spacer between cursor position and spacer
        sep_cursor_space = QFrame()
        sep_cursor_space.setFrameShape(QFrame.Shape.VLine)
        sep_cursor_space.setFrameShadow(QFrame.Shadow.Sunken)
        self.status_bar.addWidget(sep_cursor_space)

        # Big expanding spacer
        spacer_status = QWidget()
        spacer_status.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        self.status_bar.addWidget(spacer_status)

        # Progress bar
        self.progress_bar = QProgressBar()
        self.progress_bar.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.progress_bar.setVisible(False)
        self.status_bar.addPermanentWidget(self.progress_bar)

        # Separator before version and verifier
        sep_group2 = QFrame()
        sep_group2.setFrameShape(QFrame.Shape.VLine)
        sep_group2.setFrameShadow(QFrame.Shadow.Sunken)
        self.status_bar.addPermanentWidget(sep_group2)
        
        # Version label
        self.gui_version_label = QLabel(f"VSE Version: {RELEASE_VERSION}")
        self.gui_version_label.setContentsMargins(0, 0, 0, 0)
        self.status_bar.addPermanentWidget(self.gui_version_label)

        # Separator between VSE version and vehicle version
        sep_vse_verifier = QFrame()
        sep_vse_verifier.setFrameShape(QFrame.Shape.VLine)
        sep_vse_verifier.setFrameShadow(QFrame.Shadow.Sunken)
        self.status_bar.addPermanentWidget(sep_vse_verifier)
        
        self.vehicle_version_label = QLabel(f"Vehicle Version: {VERSION}")
        self.vehicle_version_label.setContentsMargins(0, 0, 0, 0)
        self.status_bar.addPermanentWidget(self.vehicle_version_label)

        # Separator between version and verifier
        sep_verifier = QFrame()
        sep_verifier.setFrameShape(QFrame.Shape.VLine)
        sep_verifier.setFrameShadow(QFrame.Shadow.Sunken)
        self.status_bar.addPermanentWidget(sep_verifier)

        # Verifier dropdown
        self.verifier_dropdown = QComboBox()
        self.verifier_dropdown.setMinimumWidth(150)
        self.verifier_dropdown.setMaximumWidth(200)
        self.verifier_dropdown.currentTextChanged.connect(self._on_verifier_selected)
        self.status_bar.addPermanentWidget(self.verifier_dropdown)
        self.verifier_paths = set() 

        # Populate verifier dropdown
        self._populate_verifier_dropdown()
        
        # Connect cursor movements to update the position indicator
        self.editor.cursorPositionChanged.connect(self.update_cursor_position)

    # --- File Operations ---

    def new_file(self):
        self.editor.clear()
        self.query_tab.clear()
        self.file_path_label.setText("No file opened")
        self.status_bar.showMessage("New file created", 3000)
        self.vcl_path = None
        self.vcl_bindings.clear()
        self.input_view.clear_input_boxes()

    def open_file(self):
        file_path, _ = QFileDialog.getOpenFileName(
            self, "Open Vehicle Specification", "", "VCL Files (*.vcl);;All Files (*)"
        )
        if not file_path:
            return
        try:
            with open(file_path, 'r', encoding='utf-8') as file:
                self.editor.setPlainText(file.read())
            self.status_bar.showMessage(f"Opened: {file_path}", 3000)
            self.set_vcl_path(file_path)

            if not self.is_valid_vcl():
                return
            self.input_view.load_inputs(self.vcl_bindings)               # Regenerate resource inputs and properties

        except Exception as e: 
            QMessageBox.critical(self, "Open File Error", f"Could not open file: {e}")
            self.file_path_label.setText("Error opening file")
            self.input_view.clear_input_boxes()         # Clear resources if file fails to load

        # Load properties
        self.regenerate_properties()
        self.update_counter_example_modes()
    
    def save_file(self):
        current_file_path = self.vcl_path

        # If no path, it's a "Save As"
        if not current_file_path:      
            file_path, _ = QFileDialog.getSaveFileName(
                self, "Save Vehicle Specification", "", "VCL Files (*.vcl);;All Files (*)"
            )
            if not file_path:
                return False
            current_file_path = file_path
        try:
            with open(current_file_path, 'w', encoding='utf-8') as file:
                file.write(self.editor.toPlainText())
            self.status_bar.showMessage(f"Saved: {current_file_path}", 3000)
            self.set_vcl_path(current_file_path)
            self.editor.document().setModified(False) 
        
        except Exception as e:
            QMessageBox.critical(self, "Save File Error", f"Could not save file: {e}")
            self.append_to_problems(f"Error saving file: {e}")
            self.input_view.clear_input_boxes()
            return False   
        
        if not self.is_valid_vcl():
            self.input_view.clear_input_boxes()
            return False
        
        # Preserve current resource state before regenerating
        self.regenerate_input_boxes()
        self.update_counter_example_modes()

        # Remember selected properties
        selected_properties = self.property_view.selected_properties()
        self.regenerate_properties(selected_properties)
        return True                            

    def save_before_operation(self):
        if not self.vcl_path or self.editor.document().isModified():
            if not self.vcl_path:
                 msg = "The file needs to be saved before this operation. Save now?"
            else:
                 msg = "The file has been modified. Save changes before this operation?"
            
            reply = QMessageBox.question(
                self, "Save File", msg,
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No | QMessageBox.StandardButton.Cancel
            )
            return reply == QMessageBox.StandardButton.Yes and self.save_file()
        return self.save_file()
    
    # --- Verifier Management ---

    def load_verifier_from_file(self):
        file_path, _ = QFileDialog.getOpenFileName(
            self, "Select Marabou Verifier", "", "Marabou Verifier (Marabou*);;All Files (*)"
        )
        if not file_path:
            return
        
        if file_path not in self.verifier_paths:
            insert_index = self.verifier_dropdown.count() - 1
            self.verifier_dropdown.insertItem(insert_index, file_path)
        
        self.verifier_paths.add(file_path)
        self.verifier_dropdown.setCurrentText(file_path)
        self._set_verifier_path(file_path)

    def _populate_verifier_dropdown(self):
        """Populate the verifier dropdown with detected verifiers."""
        self.verifier_dropdown.clear()
        verifier_name = "Marabou"

        paths = which_all(verifier_name)
        
        for path in paths:
            self.verifier_dropdown.addItem(path)
            self.verifier_paths.add(path)
        
        # If verifiers found, set the first one as default
        if self.verifier_paths:
            first_verifier_path = self.verifier_paths.pop()
            self.verifier_dropdown.setCurrentText(first_verifier_path)
            self._set_verifier_path(first_verifier_path)
        else:
            self.verifier_dropdown.addItem("No Verifier Set")
            self.verifier_dropdown.setCurrentText("No Verifier Set")
            self.verify_button.setEnabled(False)

        # Add option to load from path
        self.verifier_dropdown.addItem("Load from Path...")

    def _on_verifier_selected(self, text: str):
        """Handle verifier selection from dropdown."""
        if text == "No Verifier Set":
            self.verify_button.setEnabled(False)
        elif text == "Load from Path...":
            self.load_verifier_from_file()
        elif text in self.verifier_paths:
            self._set_verifier_path(text)
        else:
            self.verifier_dropdown.setCurrentText("No Verifier Set")

    def _set_verifier_path(self, path: str):
        """Set the verifier path and update UI."""
        self.vcl_bindings.verifier_path = path
        self.verify_button.setEnabled(True)

    # --- Compilation and Verification ---

    @pyqtSlot(str, str)
    def _gui_process_output_chunk(self, tag: str, chunk: str):
        """Processes output chunks received from the worker thread."""
        if tag == "stderr":
            current_widget = self.problems_console
        elif tag == "stdout":
            current_widget = self.log_console
        # If verifying, parse JSON chunk for progress
        if self.current_operation == 'verify' and tag == 'stdout':
            self._process_json_chunk(chunk)
        current_widget.insertPlainText(chunk)
        current_widget.ensureCursorVisible()
        self.console_tab_widget.setCurrentWidget(current_widget)

        # Refresh properties in query tab after any output
        if self.current_operation == "verify":
            self.counter_example_tab.refresh_from_cache()

    @pyqtSlot(int)
    def _gui_operation_finished(self, return_code: int):
        """Handles the completion of a VCL operation from the worker thread."""
        self.progress_bar.setVisible(False)  # Hide progress bar when operation completes
        self.compile_button.setEnabled(True)
        self.verify_button.setEnabled(True)
        self.stop_button.setVisible(False)
        self.stop_button.setEnabled(False)

        if return_code == 0: # Success
            self.status_bar.showMessage(f"{self.current_operation.capitalize()} completed successfully.", 5000)
            self.append_to_log(f"\n--- {self.current_operation.capitalize()} finished successfully. ---")
        elif return_code == -1: # Stopped by user
            self.status_bar.showMessage(f"{self.current_operation.capitalize()} stopped by user.", 5000)
            self.append_to_log(f"\n--- {self.current_operation.capitalize()} stopped by user. ---")
            if self.console_tab_widget: self.console_tab_widget.setCurrentWidget(self.problems_console)
        else: # Error
            self.status_bar.showMessage(f"Error during {self.current_operation.capitalize()}. Check Problems tab.", 5000)
            self.append_to_problems(f"\n--- {self.current_operation.capitalize()} failed. ---")
            if self.console_tab_widget: self.console_tab_widget.setCurrentWidget(self.problems_console)
        
        self.query_tab.refresh_properties()
        if self.current_operation == "verify":
            # Wait a moment to ensure all files are written
            QTimer.singleShot(500, self.counter_example_tab.refresh_from_cache)
        self.current_operation = None

    def stop_current_operation(self):
        self.status_bar.showMessage(f"Attempting to stop {self.current_operation}...", 0)
        self.stop_event.set() 
        self.stop_button.setEnabled(False) 
    
    def _start_vcl_operation(self, operation_name: str):
        """Common logic to start a VCL compile or verify operation."""
        if not self.save_before_operation():
            return

        self.input_view.assign_inputs(self.vcl_bindings)
        loaded_resources = self.input_view.get_loaded_inputs()
        if not all(loaded_resources):
            QMessageBox.warning(self, "Resource Error", "Please load all required resources (networks, datasets, parameters) before compilation/verification")
            return
        
        if operation_name == "verify":
            selected_properties = self.property_view.selected_properties()
            self.vcl_bindings.set_properties(selected_properties)
            if not selected_properties:
                QMessageBox.warning(self, "No Properties Selected", "Please select at least one property to verify.")
                return
            # Reset progress tracking
            self._json_buffer = ""
            self.verify_total_queries = 0
            self.verify_completed_queries = 0
            self.progress_bar.setValue(0)
            self.progress_bar.setMaximum(0)

        self.status_bar.showMessage(f"Performing {operation_name.capitalize()}... Please wait.", 0)
        self.compile_button.setEnabled(False)
        self.verify_button.setEnabled(False)
        self.stop_button.setEnabled(True)
        self.current_operation = operation_name
        # Show and reset progress bar for operation
        self.progress_bar.reset()
        self.progress_bar.setVisible(True)

        self.stop_event.clear() # Clear any previous stop signal
        self.problems_console.clear()

        # Set up the operation callback and reset progress for verify
        if operation_name == "compile":
            operation = functools.partial(self.vcl_bindings.compile, stop_event=self.stop_event)
        elif operation_name == "verify":
            # Reset per-property progress tracking
            self._json_buffer = ""
            self.current_property_name = ""
            self.property_total_queries = 0
            self.property_completed_queries = 0
            self.progress_bar.reset()
            operation = functools.partial(self.vcl_bindings.verify, stop_event=self.stop_event)
        else:
            return
        # Create and start worker
        worker = OperationWorker(
            operation=operation,
            vcl_bindings=self.vcl_bindings,
            stop_event=self.stop_event,
            signals=self.operation_signals
        )
        self.thread_pool.start(worker)

    def compile_spec(self):
        # Recursively clear cache directory
        for root, _, files in os.walk(CACHE_DIR, topdown=False):
            for name in files:
                try:
                    os.remove(os.path.join(root, name))
                except Exception as e:
                    self.append_to_problems(f"Error clearing cache file {name}: {e}")
        self.query_tab.refresh_properties()
        self._start_vcl_operation("compile")

    def verify_spec(self):
        self.compile_spec() # Always compile before verifying

        while self.current_operation == "compile":
            QApplication.processEvents() # Wait for compilation to finish

        if not self.vcl_bindings.verifier_path:
            QMessageBox.warning(self, "Verification Error", "Please set the verifier path first.")
            return
        self._start_vcl_operation("verify")

    # --- Resource Management ---

    def show_version(self):
        QMessageBox.information(self, "Version", f"Current version: {VERSION}")

    def regenerate_input_boxes(self):
        """Regenerate input boxes, preserving any already-set values."""
        self.input_view.load_inputs(self.vcl_bindings)

    def regenerate_properties(self, selected_properties=None):
        """Regenerate property selector, preserving any previously selected values."""
        try:
            properties = self.vcl_bindings.properties()
            self.property_view.load_properties(properties)

            if selected_properties is not None:
                for prop_name, prop_widget in self.property_view.property_items:
                    # By default, all items are checked. Uncheck those not in selected_properties
                    if prop_name not in selected_properties:
                        prop_widget.set_checked(False)
        except Exception as e:
            tb_str = traceback.format_exc()
            self.append_to_problems(f"Error loading properties: {e}\n{tb_str}")
            self.console_tab_widget.setCurrentWidget(self.problems_console)

    # --- Type Checking ---

    def is_valid_vcl(self):
        self.problems_console.clear()
        self.editor.clear_errors()
        errors = self.vcl_bindings.type_check()
        if errors:
            for error in errors:
                self.append_to_problems(f"Problem: {error['problem']}")
                self.append_to_problems(f"Fix: {error['fix']}")
                self.append_to_problems(f"Provenance: {error['provenance']['contents']}")
            self.editor.add_errors(errors)
            return False
        return True

    # --- Utility Methods ---

    def set_vcl_path(self, path):
        self.vcl_bindings.clear() # Clear any old bindings/data
        self.vcl_bindings.vcl_path = path
        self.vcl_path = path
        self.file_path_label.setText(f"File: {os.path.basename(path)}")
        self.query_tab.clear() # Clear previous output for new file

    def update_cursor_position(self):
        cursor = self.editor.textCursor()
        line = cursor.blockNumber() + 1
        col = cursor.positionInBlock() + 1
        self.position_label.setText(f"Ln {line}, Col {col}")

    def closeEvent(self, event):
        if self.current_operation:
            reply = QMessageBox.question(self, "Confirm Exit",
                                         f"A '{self.current_operation}' operation is in progress. "
                                         "Stopping it and exiting might take a moment. Exit anyway?",
                                         QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                                         QMessageBox.StandardButton.No)
            if reply == QMessageBox.StandardButton.Yes:
                if self.stop_event:
                    self.stop_event.set() # Signal stop
                self.thread_pool.waitForDone(3000) # Wait up to 3 seconds
                self.thread_pool.clear() # Clear the thread pool
                event.accept()
            else:
                event.ignore()
                return
        else:
            event.accept()

        if QApplication.instance():
            QApplication.instance().quit()

    def append_to_log(self, message: str):
        self.log_console.append(message)
        self.log_console.ensureCursorVisible()
        self.console_tab_widget.setCurrentWidget(self.log_console)
    
    def append_to_problems(self, message: str):
        self.problems_console.append(message)
        self.problems_console.ensureCursorVisible()
        self.console_tab_widget.setCurrentWidget(self.problems_console)

    def update_counter_example_modes(self):
        """Update counter example modes based on variables from properties."""
        modes = {}
        try:
            variable_renderers = self.property_view.get_variable_renderers()
            for key, renderer_class in variable_renderers.items():
                if renderer_class:
                    try:
                        # Instantiate the renderer class
                        renderer_instance = renderer_class()
                        modes[key] = renderer_instance
                    except Exception as e:
                        print(f"Error instantiating renderer {renderer_class.__name__} for {key}: {e}")
                        self.append_to_problems(f"Failed to instantiate renderer for {key}: {e}")
        except Exception as e:
            print(f"Error updating counter example modes: {e}")
            self.append_to_problems(f"Error updating counter example modes: {e}")
        self.counter_example_tab.set_modes(modes)

    def _process_json_chunk(self, chunk: str):
        """Accumulate and parse complete JSON objects from chunked output."""
        # Append new data to buffer
        self._json_buffer += chunk
        """
        Accumulate and parse complete JSON objects from chunked output using brace matching.
        """
        buf = self._json_buffer
        pos = 0
        length = len(buf)
        # Find JSON objects by matching braces
        start = buf.find('{', pos)
        while start != -1 and pos < length:
            brace_count = 0
            end = start
            for i in range(start, length):
                if buf[i] == '{':
                    brace_count += 1
                elif buf[i] == '}':
                    brace_count -= 1
                if brace_count == 0:
                    end = i
                    candidate = buf[start:end+1]
                    try:
                        obj = json.loads(candidate)
                        self._handle_json_event(obj)
                    except ValueError:
                        # Incomplete or invalid JSON
                        pass
                    pos = end + 1
                    break
            else:
                # No complete JSON object found
                break
            start = buf.find('{', pos)
        # Retain unprocessed part of buffer
        self._json_buffer = buf[pos:]

    def _init_property(self, prop: str, total: int):
        """Initialize progress bar for a new property."""
        self.current_property_name = prop
        self.property_total_queries = total
        self.property_completed_queries = 0
        self.progress_bar.setMaximum(total)
        self.progress_bar.setValue(0)
        self.progress_bar.setFormat(f"{prop}: %v/%m (%p%)")

    def _end_property_transition(self):
        """End transition pause and allow next property start."""
        self._waiting_transition = False

    def _handle_json_event(self, event: dict):
        """Update progress bar based on JSON verification events."""
        tag = event.get('tag')
        # Handle per-property progress
        if tag == 'PropertyStart':
            contents = event.get('contents', {})
            prop = contents.get('propertyName', '')
            total = contents.get('numberOfQueries', 0)
            # Schedule or perform initialization based on transition state
            def init_prop():
                self._init_property(prop, total)
            if self._waiting_transition:
                QTimer.singleShot(self.property_pause_ms, init_prop)
            else:
                init_prop()
        elif tag == 'QueryStart':
            # Optionally show halfway progress before query completes
            if self.property_total_queries > 0:
                half_value = math.ceil(self.property_completed_queries + 0.5)
                self.progress_bar.setValue(half_value)
        elif tag == 'QueryFinish':
            # Increment completed queries
            self.property_completed_queries += 1
            self.progress_bar.setValue(self.property_completed_queries)
        elif tag == 'PropertyFinish':
            # Ensure progress is complete for this property
            self.progress_bar.setValue(self.property_total_queries)
            # Enter transition pause before next property
            self._waiting_transition = True
            QTimer.singleShot(self.property_pause_ms, self._end_property_transition)


"""
Module socket_graphics.py

This module contains the graphical representation of a Socket element

Original author: Andrea Gimelli, Giacomo Rosato, Stefano Demarchi

"""

from PyQt6.QtCore import QRectF
from PyQt6.QtGui import QColor, QPen, QBrush, QPainter
from PyQt6.QtWidgets import QGraphicsItem, QStyleOptionGraphicsItem

import vehicle_gui.query_view.styling.dimension as dim
import vehicle_gui.query_view.styling.palette as palette
from ..base_types import Socket


class SocketGraphics(QGraphicsItem):
    """Graphics representation of a Socket domain model"""

    def __init__(self, socket: 'Socket', parent=None):
        super().__init__(parent)
        # Reference to socket domain model
        self.socket_ref = socket

        # Style parameters
        self.bg_color = QColor(palette.WHITE)
        self.outline_color = QColor(palette.WHITE)
        self.init_colors()

        self._pen = QPen(self.outline_color)
        self._pen.setWidth(dim.SOCKET_OUTLINE)
        self._brush = QBrush(self.bg_color)

        self.setZValue(-1)

    def init_colors(self):
        """Set socket colors using polymorphic behavior from the block"""
        socket_colors = self.socket_ref.block_ref.get_socket_colors()
        self.bg_color = QColor(socket_colors["bg_color"])
        self.outline_color = QColor(socket_colors["outline_color"])

    def paint(self, painter: QPainter, option: 'QStyleOptionGraphicsItem', widget=None) -> None:
        painter.setBrush(self._brush)
        painter.setPen(self._pen)
        painter.drawEllipse(-dim.SOCKET_RADIUS, -dim.SOCKET_RADIUS, 2 * dim.SOCKET_RADIUS, 2 * dim.SOCKET_RADIUS)

    def boundingRect(self) -> QRectF:

        return QRectF(
            - dim.SOCKET_RADIUS - dim.SOCKET_OUTLINE,
            - dim.SOCKET_RADIUS - dim.SOCKET_OUTLINE,
            2 * (dim.SOCKET_RADIUS + dim.SOCKET_OUTLINE),
            2 * (dim.SOCKET_RADIUS + dim.SOCKET_OUTLINE),
        )



"""
Module base_types.py

This module contains the base classes for the node editor framework.

"""

from typing import Any
from enum import Enum
from .styling import dimension as dim


# Core enumerations
class SocketType(Enum):
    """Socket type enumeration"""
    INPUT = 0
    OUTPUT = 1


# Base classes
class Block:
    """Base block class for all node types"""
    def __init__(self):
        self.id = None
        self.title = "Block"    
        self.scene_ref = None           # Reference to the scene this block belongs to
        self.graphics = None            # Reference to BlockGraphics instance
        self.inputs = []                # Input sockets
        self.outputs = []               # Output sockets
        self.x = 0                      
        self.y = 0

    def has_parameters(self):
        """Required method for GraphicsBlock"""
        return False
    
    def get_color_scheme(self):
        """Return the color scheme for this block type"""
        # Default color scheme - subclasses should override
        return ["#304153", "#34475b", "#34475b", "#304153", "#2c2c2c"]
    
    def get_socket_colors(self):
        """Return the socket colors for this block type"""
        # Default socket colors - subclasses should override  
        return {"bg_color": "#FFFFFF", "outline_color": "#FFFFFF"}
    
    def get_block_width(self):
        """Return the width for this block type"""
        # Default width - subclasses can override
        return dim.BLOCK_BASE_WIDTH
    
    def update_edges(self):
        """Update connected edges when block moves"""
        for socket in self.sockets:
            for edge in socket.edges:   
                edge.update_graphics_position()
        
        # Also update any edges in the scene that involve this block
        if self.scene_ref:
            for edge in self.scene_ref.edges.values():
                if (edge.start_skt.block_ref == self or edge.end_skt.block_ref == self):
                    edge.update_graphics_position()


class Socket:
    """Base socket class for block connections"""
    def __init__(self, socket_type, block):
        self.s_type = socket_type
        self.block_ref = block
        self.graphics = None  # Reference to SocketGraphics instance
        self.socket_graphics = None  # Legacy reference to SocketGraphics instance
        self.edges = []  # List of connected edges


class Edge:
    """Base edge class for connecting blocks"""
    def __init__(self, start_socket, end_socket):
        self.id = id(self)
        self.start_skt = start_socket
        self.end_skt = end_socket
        self.view_dim = True
        self.graphics = None  # Reference to EdgeGraphics instance
        self.edge_graphics = None  # Legacy reference to EdgeGraphics instance
        
        # Add this edge to the sockets
        start_socket.edges.append(self)
        end_socket.edges.append(self)
    
    def update_graphics_position(self):
        """Update the graphics position of this edge"""
        if self.graphics and self.start_skt.block_ref.graphics and self.end_skt.block_ref.graphics:
            start_pos = self.start_skt.block_ref.graphics.pos()
            end_pos = self.end_skt.block_ref.graphics.pos()
            start_graphics = self.start_skt.block_ref.graphics
            end_graphics = self.end_skt.block_ref.graphics
            
            # Source socket is at bottom center of source block
            start_x = start_pos.x() + start_graphics.width / 2
            start_y = start_pos.y() + start_graphics.height
            
            # Target socket is at top center of target block  
            end_x = end_pos.x() + end_graphics.width / 2
            end_y = end_pos.y()
            
            self.graphics.src_pos = [start_x, start_y]
            self.graphics.dest_pos = [end_x, end_y]
            self.graphics.update_path()
            
            # Force graphics updates to ensure the edge is rendered correctly
            self.graphics.update()
            self.graphics.prepareGeometryChange()


class Scene:
    """Base scene class for managing blocks and edges"""
    def __init__(self):
        self.blocks = {}
        self.edges = {}
        self.counter = 0
    
    def add_block(self, block):
        """Add a block to the scene"""
        block.id = self.counter
        block.scene_ref = self
        self.blocks[self.counter] = block
        self.counter += 1
        return block
    
    def connect_blocks(self, source_block, target_block):
        """Connect two blocks in the verification workflow"""
        source_socket = next((s for s in source_block.sockets if s.s_type == SocketType.OUTPUT), None)
        target_socket = next((s for s in target_block.sockets if s.s_type == SocketType.INPUT), None)
        
        if source_socket and target_socket:
            edge = Edge(source_socket, target_socket)
            self.edges[edge.id] = edge
            return edge
        return None


# Utility functions
def get_classname(obj: Any) -> str:
    """Get the class name of an object"""
    return obj.__class__.__name__



# File generated by the BNF Converter (bnfc 2.9.5).

import pygments.lexer
from pygments.token import *
__all__ = ["ExternalLexer"]
class ExternalLexer(pygments.lexer.RegexLexer):
    name = 'External'
    aliases = ['external']
    KEYWORDS = ['in', 'record', 'type', 'where']
    def get_tokens_unprocessed(self, text):
        for index, token, value in super(ExternalLexer,self).get_tokens_unprocessed(text):
            if token is Name and value in self.KEYWORDS:
                yield index, Keyword, value
            else:
                yield index, token, value
    tokens = {
        'root': [
            (r'--.*\n', Comment),
            (r'\{-((.)(?<!-))*-((.)(?<![-\}])((.)(?<!-))*-|-)*\}', Comment),
            (r'True|False', Name),
            (r'(\d)+', Name),
            (r'(\d)+\.(\d)+', Name),
            (r'@(network|dataset|parameter|property|tensor)', Name),
            (r'@postulate', Name),
            (r'\{', Name),
            (r'\}', Name),
            (r'->', Name),
            (r'forallT', Name),
            (r'if', Name),
            (r'then', Name),
            (r'else', Name),
            (r'\.', Name),
            (r':', Name),
            (r'\\', Name),
            (r'let', Name),
            (r'Type', Name),
            (r'Unit', Name),
            (r'Bool', Name),
            (r'Index', Name),
            (r'Nat', Name),
            (r'Real', Name),
            (r'List', Name),
            (r'Vector', Name),
            (r'Tensor', Name),
            (r'forall', Name),
            (r'exists', Name),
            (r'foreach', Name),
            (r'=>', Name),
            (r'and', Name),
            (r'or', Name),
            (r'not', Name),
            (r'==', Name),
            (r'!=', Name),
            (r'<=', Name),
            (r'<', Name),
            (r'>=', Name),
            (r'>', Name),
            (r'==\.', Name),
            (r'!=\.', Name),
            (r'<=\.', Name),
            (r'<\.', Name),
            (r'>=\.', Name),
            (r'>\.', Name),
            (r'\*', Name),
            (r'/', Name),
            (r'\+', Name),
            (r'-', Name),
            (r'min', Name),
            (r'max', Name),
            (r'nil', Name),
            (r'::', Name),
            (r'\[', Name),
            (r'\]', Name),
            (r'!', Name),
            (r'map', Name),
            (r'fold', Name),
            (r'reduceOr', Name),
            (r'reduceAnd', Name),
            (r'reduceAdd', Name),
            (r'reduceMul', Name),
            (r'reduceMin', Name),
            (r'reduceMax', Name),
            (r'HasLeq', Name),
            (r'HasEq', Name),
            (r'HasNotEq', Name),
            (r'HasAdd', Name),
            (r'HasSub', Name),
            (r'HasMul', Name),
            (r'HasFold', Name),
            (r'HasMap', Name),
            (r'IsTensorType', Name),
            (r'name', Name),
            (r'infer', Name),
            (r'[a-zA-Z](_|\d|[a-zA-Z])*', Name),
            (r'\?(_|\d|[a-zA-Z])*', Name),
            (r'\.[a-zA-Z](_|\d|[a-zA-Z])*', Name),
            (r'[a-zA-Z]([a-zA-Z]|\d|_|\')*', Name),
            (r'@0|\(|\)|\{|\}|\{\{|\}\}|=|,|\(\)|;', Operator),
            (r'(\d)+', Number.Integer),
            (r'(\d)+\.(\d)+(e(-)?(\d)+)?', Number.Float),
            (r'"((.)(?<!["\\])|\\["\\nt])*"', String.Double),
            (r'\'((.)(?<![\'\\])|\\[\'\\nt])\'', String.Char),
            (r'\s+', Token.Space)
        ]
    }



from abc import ABC, abstractmethod
from PyQt6.QtWidgets import QWidget, QLabel, QVBoxLayout, QTextEdit
from PyQt6.QtGui import QPixmap, QImage
from typing import Optional, Dict
import numpy as np


class BaseRenderer(ABC):
    @abstractmethod
    def render(self, data: np.ndarray):
        """Render the given data in the internal widget."""
        pass

    @property
    @abstractmethod
    def widget(self):
        """Subclasses must provide a widget property."""
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        """Subclasses must provide a name property."""
        pass


class GSImageRenderer(BaseRenderer):
    def __init__(self):
        self._widget = QLabel()

    def _prepare_pixmap(self, data: np.ndarray) -> QPixmap:
        h, w = data.shape
        qimage = QImage(data.copy().data, w, h, w, QImage.Format_Grayscale8)
        return QPixmap.fromImage(qimage)

    def render(self, data: np.ndarray):
        pixmap = self._prepare_pixmap(data)
        self.widget.setPixmap(pixmap)

    @property
    def widget(self) -> QWidget:
        return self._widget
    
    @property
    def name(self) -> str:
        return "Grayscale Image Renderer"

class TextRenderer(BaseRenderer):
    """Renderer for text data."""
    def __init__(self):
        self._widget = QTextEdit()
        self._widget.setReadOnly(True)

    def render(self, data: np.ndarray):
        with np.printoptions(precision=3, threshold=20, suppress=True):
            string_rept = str(data)
            self.widget.setText(string_rept)

    @property
    def widget(self) -> QWidget:
        return self._widget
    
    @property
    def name(self) -> str:
        return "Text Renderer"



import os
from PyQt6.QtCore import Qt, QSize, pyqtSignal
from PyQt6.QtGui import QFont, QFontDatabase, QIcon
from PyQt6.QtWidgets import QVBoxLayout, QPushButton, QLabel, QLineEdit, QFrame, QFileDialog, QMessageBox, QSizePolicy, QHBoxLayout, QWidget, QScrollArea


class InputBox(QFrame):
    # Signal emitted when load status changes
    load_status_changed = pyqtSignal(bool)  # True when loaded, False when unloaded

    def __init__(self, name, type_, data_type=None):
        super().__init__()
        self.setObjectName("InputBox")
        layout = QVBoxLayout()
        title = QLabel(f"{type_}: {name}")
        mono = QFontDatabase.systemFont(QFontDatabase.FixedFont)
        mono.setPointSize(11)
        mono.setWeight(QFont.Weight.Bold)
        title.setFont(mono)
        layout.addWidget(title)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.is_loaded = False
        self.type = type_
        self.name = name
        self.data_type = data_type

        if type_ == "Network" or type_ == "Dataset" or type_ == "Variable":
            input_layout = QHBoxLayout()
            self.input_box = QLineEdit()
            self.input_box.setPlaceholderText(f"No {type_} loaded")
            self.input_box.setReadOnly(True)
            input_layout.addWidget(self.input_box)

            # Create small folder icon button
            self.load_btn = QPushButton()
            self.load_btn.setIcon(QIcon.fromTheme("folder"))
            self.load_btn.setFixedSize(32, 32)
            self.load_btn.clicked.connect(self.set_path)
            input_layout.addWidget(self.load_btn)
            input_layout.setSpacing(4)
            layout.addLayout(input_layout)

        elif type_ == "Parameter":
            self.value = None       # Unset value is None

            self.input_box = QLineEdit()
            self.input_box.editingFinished.connect(self.set_value)
            layout.addWidget(self.input_box)

            # Add a label to show the data type
            self.data_type_label = QLabel(f"Data Type: {data_type}")
            label_font = QFontDatabase.systemFont(QFontDatabase.FixedFont)
            label_font.setPointSize(10)
            self.data_type_label.setFont(label_font)
            self.data_type_label.setAlignment(Qt.AlignmentFlag.AlignRight)
            layout.addWidget(self.data_type_label)
            self.input_box.setPlaceholderText(f"Enter {self.data_type} value")
            self.data_type = data_type

        self.setLayout(layout)

    def set_path(self):
        """Open a dataset path"""
        if self.type == "Network":
            file_filter = "ONNX Files (*.onnx);;All Files (*)"
        elif self.type == "Dataset":
            file_filter = "All Files (*)"
        elif self.type == "Variable":
            file_filter = "Renderer modules (*.py);;All Files (*)"

        file_path, _ = QFileDialog.getOpenFileName(
            self, f"Open {self.type}", "", file_filter
        )
        if not file_path:
            return
        self.path = file_path
        self.input_box.setText(os.path.basename(file_path))
        self.input_box.setToolTip(file_path) # Show full path on hover
        self.is_loaded = True
        self.load_status_changed.emit(True)

    def set_value(self):
        value = self.input_box.text()
        """Set the value of the parameter"""
        if self.data_type == "Real":
            try:
                value = float(value)
            except ValueError:
                QMessageBox.critical(self, "Invalid Value", "Value must be a number.")
                return
        elif self.data_type == "Nat":
            try:
                value = int(value)
            except ValueError:
                QMessageBox.critical(self, "Invalid Value", "Value must be a natural number.")
                return
        elif self.data_type == "Bool":
            if value.lower() not in ["true", "false"]:
                QMessageBox.critical(self, "Invalid Value", "Value must be 'true' or 'false'.")
                return
            value = value.lower() == "true"
        else:
            raise ValueError(f"Unexpected data type: {self.data_type}")

        self.input_box.setText(str(value))
        self.is_loaded = True
        self.value = value
        self.load_status_changed.emit(True)


class InputView(QWidget):
    """Widget for managing input boxes (Networks, Datasets, Parameters)."""

    def __init__(self, parent=None, error_callback=None):
        super().__init__(parent)
        self.error_callback = error_callback or (lambda msg: print(f"Error: {msg}"))
        self.input_boxes = []
        # Store input state: name -> {"definition": {...}, "loaded": (path/value, display_text) or None}
        self._input_state = {}

        # Create scrollable area for input boxes
        self.scroll_area = QScrollArea()
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        # Create a widget to hold the input boxes
        self.content_widget = QWidget()
        self.input_layout = QVBoxLayout(self.content_widget)
        self.input_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.scroll_area.setWidget(self.content_widget)

        # Status label
        self.status_label = QLabel("0 inputs loaded")

        # Control buttons
        btn_clear = QPushButton("Clear All")
        btn_clear.clicked.connect(self.clear_input_boxes)

        btn_row = QHBoxLayout()
        btn_row.addWidget(btn_clear)
        btn_row.addStretch()

        layout = QVBoxLayout(self)
        layout.addWidget(self.status_label)
        layout.addLayout(btn_row)
        layout.addWidget(self.scroll_area)

    def _remove_all_boxes(self):
        """Remove all input boxes without recreating them."""
        # Remove and delete existing boxes
        while self.input_layout.count():
            item = self.input_layout.takeAt(0)
            widget = item.widget()
            if widget:
                widget.deleteLater()
        self.input_boxes.clear()

    def clear_input_boxes(self):
        """Delete all input boxes and recreate empty ones."""
        self._remove_all_boxes()

        # Recreate empty boxes from stored input definitions
        for name, state in self._input_state.items():
            definition = state["definition"]
            type_ = definition.get("tag")
            data_type = definition.get("typeText", None)
            if not name or not type_:
                continue
            box = InputBox(name, type_, data_type=data_type)
            box.load_status_changed.connect(self._update_status)
            self.input_layout.addWidget(box)
            self.input_boxes.append(box)

        self._update_status()

    def _save_loaded_state(self):
        """Update the loaded state of all input boxes."""
        for box in self.input_boxes:
            if box.name in self._input_state:
                if box.is_loaded:
                    if hasattr(box, 'path'):
                        # For Network, Dataset, Variable
                        self._input_state[box.name]["loaded"] = (box.path, box.input_box.text())
                    elif hasattr(box, 'value'):
                        # For Parameter
                        self._input_state[box.name]["loaded"] = (box.value, box.input_box.text())
                else:
                    self._input_state[box.name]["loaded"] = None

    def load_inputs(self, vcl_bindings):
        """Load inputs from VCL bindings and create input boxes."""
        self._save_loaded_state()
        self._remove_all_boxes()

        try:
            inputs = vcl_bindings.resources()
            
            # Update input state with new definitions, preserving loaded state where possible
            new_input_state = {}
            for entry in inputs:
                name = entry.get("name")
                if name:
                    # Preserve existing loaded state if input still exists
                    existing_loaded = None
                    if name in self._input_state and self._input_state[name]["loaded"] is not None:
                        existing_loaded = self._input_state[name]["loaded"]
                    
                    new_input_state[name] = {
                        "definition": entry,
                        "loaded": existing_loaded
                    }
            self._input_state = new_input_state

            for name, state in self._input_state.items():
                definition = state["definition"]
                type_ = definition.get("tag")
                data_type = definition.get("typeText", None)
                if not name or not type_:
                    print(f"Skipping input entry with missing name or type: {definition}")
                    continue
                box = InputBox(name, type_, data_type=data_type)
                box.load_status_changed.connect(self._update_status)
                self.input_layout.addWidget(box)
                self.input_boxes.append(box)

                # Restore loaded state if it exists
                if state["loaded"] is not None:
                    value, display_text = state["loaded"]
                    if type_ in ["Network", "Dataset", "Variable"]:
                        box.path = value
                        box.input_box.setText(display_text)
                        box.input_box.setToolTip(value)  # Show full path on hover
                        box.is_loaded = True
                    elif type_ == "Parameter":
                        box.value = value
                        box.input_box.setText(display_text)
                        box.is_loaded = True

        except Exception as e:
            import traceback
            tb_str = traceback.format_exc()
            self.error_callback(f"Error generating input boxes: {e}\n{tb_str}")

        self._update_status()

    def assign_inputs(self, vcl_bindings):
        """Assign inputs from InputBox widgets to the VCLBindings object."""
        assigned = []
        for box in self.input_boxes:
            if box.is_loaded:
                try:
                    if box.type == "Network":
                        vcl_bindings.set_network(box.name, box.path)
                    elif box.type == "Dataset":
                        vcl_bindings.set_dataset(box.name, box.path)
                    elif box.type == "Parameter":
                        vcl_bindings.set_parameter(box.name, box.value)
                    assigned.append(box.name)
                except Exception as e:
                    self.error_callback(f"Error assigning input {box.name}: {e}")
        return assigned

    def regenerate_input_boxes(self):
        """Regenerate input boxes, preserving any previously loaded values."""
        # Restore values from stored loaded state to the newly created boxes
        for box in self.input_boxes:
            if box.name in self._loaded_state:
                box.is_loaded = True
                state_data = self._loaded_state[box.name]
                if box.type in ["Network", "Dataset"]:
                    box.path, display_text = state_data
                    box.input_box.setText(display_text)
                    box.input_box.setToolTip(box.path)
                elif box.type == "Parameter":
                    box.value = state_data
                    box.input_box.setText(str(state_data))

        self._update_status()

    def get_loaded_inputs(self):
        """Get list of loaded input names"""
        return [box.name for box in self.input_boxes if box.is_loaded]

    def _update_status(self):
        """Update the status label with current input count."""
        loaded = len([box for box in self.input_boxes if box.is_loaded])
        total = len(self.input_boxes)
        self.status_label.setText(f"{loaded} / {total} inputs loaded")


import importlib.util
import inspect
from typing import Type
from pathlib import Path
from vehicle_gui.counter_example_view.base_renderer import BaseRenderer


def load_renderer_classes(path: str) -> list[Type[BaseRenderer]]:
    """Load renderer class types from a Python file without instantiating them."""
    path = Path(path)
    spec = importlib.util.spec_from_file_location(path.stem, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    renderer_classes = []

    for _, obj in inspect.getmembers(module, inspect.isclass):
        if issubclass(obj, BaseRenderer) and obj is not BaseRenderer:
            renderer_classes.append(obj)

    return renderer_classes


import os
import re
import sys
import vehicle_lang as vcl
import json
import asyncio
from vehicle_lang import VehicleError
from typing import Sequence, Optional, Callable
from vehicle_gui.vcl_utils import get_resources_info
from vehicle_gui.vcl_utils import get_properties_info
from vehicle_gui import VEHICLE_DIR

CACHE_DIR = os.path.join(VEHICLE_DIR, "cache")


class Runner:
	def __init__(self, command: str,  script: str = "_run_vcl.py", *args: str, **kwargs: str):
		self.script_path = os.path.join(os.path.dirname(__file__), script)
		self.cmd = self.build_command(command, args, kwargs)

	def build_command(self, command: str, args: Sequence[str], kwargs: dict) -> list[str]:
		cmd = [sys.executable, "-u", self.script_path, command]
		cmd.extend(args)

		# Keyword is --option
		for option, value in kwargs.items():
			flag = f"--{option.replace('_', '-')}"
			# Handle networks, datasets, and parameters, which requires flag to be repeated for each name-value pair
			if isinstance(value, dict):
				for name, val in value.items():
					cmd.append(flag)
					cmd.append(f"{name}:{val}")
			# Handle properties, which requires flag to be repeated for each property
			elif isinstance(value, list):
				for item in value:
					cmd.append(flag)
					cmd.append(str(item))
			# Handle other parameters
			else:
				cmd.append(flag)
				cmd.append(str(value))
				
		# Always include JSON output for verify commands
		if command == "verify":
			cmd.append("--json")
		return cmd

	async def run(self, line_reader: Callable, finish_fn: Callable, stop_event: asyncio.Event) -> str:
		process = await asyncio.create_subprocess_exec(
			*self.cmd,
			stdout=asyncio.subprocess.PIPE,
			stderr=asyncio.subprocess.PIPE
		)

		async def stream_output(stream, tag):
			while True:
				data_chunk = await stream.read(4096)
				if not data_chunk:
					break
				decoded = data_chunk.decode(errors="replace")
				line_reader(tag, decoded)

		async def watch_stop():
			while True:
				await asyncio.sleep(0.1)
				if stop_event.is_set():
					print(f"[DEBUG] Stopping process: {self.cmd}")
					try:
						process.terminate()
					except ProcessLookupError:
						pass

		stdout_task = asyncio.create_task(stream_output(process.stdout, "stdout"))
		stderr_task = asyncio.create_task(stream_output(process.stderr, "stderr"))
		stop_task = asyncio.create_task(watch_stop())

		await asyncio.wait([stdout_task, stderr_task, stop_task], return_when=asyncio.FIRST_COMPLETED)

		exit_code = await process.wait()
		finish_fn(exit_code)

	def run_sync(self, command: str, *args: str, **kwargs: str) -> str:
		return asyncio.run(self.run(command, *args, **kwargs))


class VCLBindings:
	"""Python bindings for the Vehicle command-line tool"""
	def __init__(self, vcl_path = None):
		"""Initialize the VCLBindings class"""
		self._vcl_path = vcl_path
		self._verifier_path = None
		self._networks = {}
		self._datasets = {}
		self._parameters = {}
		self._properties = []

	def compile(self, callback_fn: Callable, finish_fn: Callable, stop_event: asyncio.Event):
		"""Compile a VCL specification"""
		runner = Runner(
			command="compile", 
			specification=self._vcl_path, 
			network=self._networks,
			dataset=self._datasets,
			parameter=self._parameters,
			target="MarabouQueries",
			output=CACHE_DIR,
		)
		asyncio.run(runner.run(
			line_reader=callback_fn, 
			finish_fn=finish_fn,
			stop_event=stop_event,
		))
	
	def verify(self, callback_fn: Callable, finish_fn: Callable, stop_event: asyncio.Event):	
		"""Verify a VCL specification"""	
		runner = Runner(
			command="verify", 
			specification=self._vcl_path, 
			verifier="Marabou", 
			verifier_location=self._verifier_path, 
			network=self._networks,
			dataset=self._datasets,
			parameter=self._parameters,
			cache=CACHE_DIR,
			property=self._properties,
		)
		asyncio.run(runner.run(
			line_reader=callback_fn, 
			finish_fn=finish_fn,
			stop_event=stop_event,
		))

	def type_check(self):
		"""Type check a VCL specification"""
		try:
			# Temporary method until type checking is fixed in vehicle_lang
			vcl.list(self._vcl_path)
		except VehicleError as e:
			error_str = str(e)
			error_json = json.loads(error_str)
			if "provenance" not in error_json:
				# error is not JSON formatted. This is a syntax error caught by BNFC parser
				match = re.search(r"line (\d+), column (\d+)", error_str)
				if match:
					line, column = [int(g) for g in match.groups()]
					start_col = max(1, column - 3)
					end_col = column + 2

					error_json = [{
						"provenance": {
							"tag": "Provenance",
							"contents": [line, start_col, line, end_col]
						},
						"problem": "Syntax Error",
						"fix": f"fix {error_json}"
					}]
			if not isinstance(error_json, list):
				error_json = [error_json]
			return error_json
		return []

	def resources(self):
		"""Get the resources used by the VCLBindings"""
		vcl_output = get_resources_info(self._vcl_path)
		if not vcl_output:
			raise VehicleError("No resources found in VCL file. Something is wrong.")
		return json.loads(vcl_output)

	def properties(self):
		"""Get the properties in the VCL specification"""
		vcl_output = get_properties_info(self._vcl_path)
		if not vcl_output:
			return []
		return json.loads(vcl_output)
	
	def variables(self):
		"""Get the quantified variables listed in the VCL specification"""
		result = get_properties_info(self._vcl_path)
		if not result:
			return []
		result = json.loads(result)
		vars = set()
		for item in result:
			vars = vars.union(set(item["quantifiedVariablesInfo"]))
		return [{"name": var, "tag": "Variable"} for var in vars]

	@property
	def vcl_path(self):
		"""Get the VCL path"""
		return self._vcl_path

	@vcl_path.setter
	def vcl_path(self, value):
		"""Set the VCL path"""
		# Check if the file exists
		if os.path.isfile(value):
			self._vcl_path = value
		else:
			raise FileNotFoundError(f"VCL file not found: {value}")

	@property
	def verifier_path(self):
		"""Get the VCL path"""
		return self._verifier_path

	@verifier_path.setter
	def verifier_path(self, value):
		"""Set the Marabou Verifier path"""
		# Check if the file exists
		if os.path.isfile(value):
			self._verifier_path = value
		else:
			raise FileNotFoundError(f"Verifier binary not found: {value}")

	def set_network(self, name, path):
		"""Add a network to the VCLBindings"""
		if not os.path.isfile(path):
			raise FileNotFoundError(f"Network file not found: {path}")
		self._networks[name] = path

	def set_dataset(self, name, path):
		"""Add a dataset to the VCLBindings"""
		if not os.path.isfile(path):
			raise FileNotFoundError(f"Dataset file not found: {path}")
		self._datasets[name] = path

	def set_parameter(self, name, value):
		"""Add a property to the VCLBindings"""
		self._parameters[name] = value

	def set_properties(self, properties: list[str]):
		"""Set the list of properties to verify"""
		self._properties = properties

	def clear(self):
		"""Clear all networks, datasets, and properties"""
		self._networks.clear()
		self._datasets.clear()
		self._parameters.clear()
		self._properties.clear()
		self._vcl_path = None


from pathlib import Path

from PyQt6.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QLabel, QComboBox, QLineEdit, QCheckBox, QFrame, QScrollArea
from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import QFileDialog

from vehicle_gui.counter_example_view.extract_renderers import load_renderer_classes
from vehicle_gui.counter_example_view.base_renderer import TextRenderer, GSImageRenderer
from vehicle_gui import VEHICLE_DIR

RENDERERS_DIR = Path(VEHICLE_DIR) / "renderers"


class VariableBox(QWidget):
    """Widget for variable renderer selection"""
    
    renderer_changed = pyqtSignal(str, object)  # variable_name, renderer_class
    
    def __init__(self, variable_name, parent=None):
        super().__init__(parent)
        self.variable_name = variable_name
        self.renderer_class = None
        self.renderer_map = {}  # Maps dropdown text to renderer class
        
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        
        # Variable name label
        name_label = QLabel(f"{variable_name}:")
        name_label.setMinimumWidth(80)
        name_label.setAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        layout.addWidget(name_label)
        
        # Display box (read-only line edit like resource boxes)
        self.display_box = QLineEdit()
        self.display_box.setPlaceholderText("No renderer loaded")
        self.display_box.setReadOnly(True)
        layout.addWidget(self.display_box)
        
        # Renderer selection dropdown
        self.renderer_combo = QComboBox()
        self._populate_renderer_dropdown()
        self.renderer_combo.currentTextChanged.connect(self._on_renderer_changed)
        layout.addWidget(self.renderer_combo)
        
        layout.setSpacing(4)
        self.setLayout(layout)
    
    def _populate_renderer_dropdown(self):
        """Populate the renderer dropdown with built-in and cached renderers."""
        # Add built-in renderers
        self.renderer_combo.addItem("GSImage Renderer")
        self.renderer_map["GSImage Renderer"] = GSImageRenderer
        
        self.renderer_combo.addItem("Text Renderer")
        self.renderer_map["Text Renderer"] = TextRenderer
        
        # Scan VEHICLE_DIR/renderers for additional renderers
        for py_file in RENDERERS_DIR.glob("*.py"):
            try:
                renderer_classes = load_renderer_classes(str(py_file))
                for renderer_class in renderer_classes:
                    display_name = f"{renderer_class.__name__}"
                    self.renderer_combo.addItem(display_name)
                    self.renderer_map[display_name] = renderer_class
            except Exception as e:
                error_msg = f"Error loading renderers from {py_file}: {e}"
                print(error_msg)
                raise RuntimeError(error_msg)
        
        # Add custom loader option
        self.renderer_combo.addItem("Load From Path...")
    
    def _on_renderer_changed(self, selection):
        if selection == "Load From Path...":
            self._load_from_path()
        elif selection in self.renderer_map:
            self._set_renderer_from_map(selection)
    
    def _set_renderer_from_map(self, selection):
        """Set renderer from the pre-loaded renderer map."""
        renderer_class = self.renderer_map[selection]
        try:
            self._set_renderer_class(renderer_class)
        except Exception as e:
            self._set_renderer_error(str(e))
            print(f"Error setting renderer {selection}: {e}")
    
    def _load_from_path(self):
        file_path, _ = QFileDialog.getOpenFileName(
            self, f"Load Renderer for {self.variable_name}", "", "Renderer modules (*.py);;All Files (*)"
        )
        if file_path:
            try:
                renderer_classes = load_renderer_classes(file_path)
                if renderer_classes:
                    if len(renderer_classes) > 1:
                        print(f"Warning: Multiple renderer classes found in {file_path}. Using the first one.")
                    renderer = renderer_classes[0]
                    self._set_renderer_class(renderer, f"Custom: {Path(file_path).name}", file_path, "Load Custom Renderer...")
                else:
                    self._set_renderer_error("No valid renderer class found in the selected file")
            except Exception as e:
                self._set_renderer_error(str(e))
                print(f"Error loading renderer from {file_path}: {e}")
    
    def _set_renderer_class(self, renderer_class):
        """Common method to set a renderer class and update UI."""
        display_text = renderer_class.__name__
        self.renderer_class = renderer_class
        self.display_box.setText(display_text)
        self.renderer_combo.setCurrentText(display_text)
        self.renderer_changed.emit(self.variable_name, renderer_class)
    
    def _set_renderer_error(self, error_message):
        """Common method to handle renderer loading errors."""
        self.display_box.setText("Failed to load renderer")
        self.display_box.setToolTip(error_message)


class PropertyBox(QFrame):
    """Widget representing a single property with its variables."""
    
    property_toggled = pyqtSignal(str, bool)  # property_name, checked
    
    def __init__(self, prop_data, parent=None):
        super().__init__(parent)
        self.prop_data = prop_data
        self.setFrameStyle(QFrame.Shape.Box)
        self.setLineWidth(1)
        
        layout = QVBoxLayout()
        layout.setContentsMargins(8, 8, 8, 8)
        layout.setSpacing(6)
        
        # Property checkbox and label
        prop_layout = QHBoxLayout()
        self.prop_checkbox = QCheckBox()
        self.prop_checkbox.setChecked(True)
        self.prop_checkbox.stateChanged.connect(self._on_property_toggled)
        
        prop_label = QLabel(f"{prop_data['name']} : {prop_data['type']}")
        prop_label.setWordWrap(True)
        prop_layout.addWidget(self.prop_checkbox)
        prop_layout.addWidget(prop_label, 1)
        prop_layout.addStretch()
        layout.addLayout(prop_layout)
        
        # Variables section
        variables = prop_data.get("quantifiedVariablesInfo", [])
        if variables:
            # Container for variable widgets with indentation
            self.var_container = QWidget()
            var_layout = QVBoxLayout(self.var_container)
            var_layout.setContentsMargins(20, 5, 5, 5)  # Left margin for indentation
            
            # Add a subtle separator
            separator = QFrame()
            separator.setFrameShape(QFrame.Shape.HLine)
            separator.setFrameShadow(QFrame.Shadow.Sunken)
            var_layout.addWidget(separator)
            
            var_title = QLabel("Variables:")
            var_title.setStyleSheet("font-weight: bold;")
            var_layout.addWidget(var_title)
            
            self.variable_widgets = {}
            for var_name in variables:
                var_widget = VariableBox(var_name)
                if parent and hasattr(parent, '_on_variable_renderer_changed'):
                    var_widget.renderer_changed.connect(parent._on_variable_renderer_changed)
                var_layout.addWidget(var_widget)
                self.variable_widgets[var_name] = var_widget
            
            layout.addWidget(self.var_container)
        
        self.setLayout(layout)
    
    def _on_property_toggled(self, state):
        checked = state == Qt.CheckState.Checked
        self.property_toggled.emit(self.prop_data['name'], checked)
    
    def is_checked(self):
        return self.prop_checkbox.isChecked()
    
    def set_checked(self, checked):
        self.prop_checkbox.setChecked(checked)


class PropertyView(QWidget):
    selection_changed = pyqtSignal(list)  # list of selected property ids
    renderers_changed = pyqtSignal()  # emitted when any variable renderer changes

    def __init__(self, parent=None):
        super().__init__(parent)
        self._properties = []  # {"name": str, "type": str, "quantifiedVariablesInfo": list}
        self.property_widgets = {}
        self._variable_renderers = {}
        self._is_syncing = False  # Flag to prevent recursive synchronization
        
        # Create scrollable area for properties
        self.scroll_area = QScrollArea()
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        
        self.content_widget = QWidget()
        self.content_layout = QVBoxLayout(self.content_widget)
        self.content_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.scroll_area.setWidget(self.content_widget)
        
        self.status_label = QLabel("0 / 0 selected")

        btn_all = QPushButton("All")
        btn_none = QPushButton("None")
        btn_all.clicked.connect(self.select_all)
        btn_none.clicked.connect(self.select_none)

        btn_row = QHBoxLayout()
        btn_row.addWidget(btn_all)
        btn_row.addWidget(btn_none)
        
        # Synchronization checkbox
        self.sync_checkbox = QCheckBox("Sync variables with same name")
        self.sync_checkbox.setToolTip("When enabled, changing a renderer for a variable will apply to all variables with the same name")
        btn_row.addWidget(self.sync_checkbox)
        
        btn_row.addStretch()

        layout = QVBoxLayout(self)
        layout.addWidget(self.status_label)
        layout.addLayout(btn_row)
        layout.addWidget(self.scroll_area)

    def load_properties(self, props):
        """
        Loads a list of properties into the widget. Properties are a list of dictionaries in the form:
        {"name": str, "type": str, "quantifiedVariablesInfo": list}
        """
        self._properties = props
        self.property_widgets = {}
        
        # Clear existing widgets
        while self.content_layout.count():
            item = self.content_layout.takeAt(0)
            if item.widget():
                item.widget().deleteLater()
        
        # Create property widgets
        for prop in props:
            prop_widget = PropertyBox(prop, self)
            prop_widget.property_toggled.connect(self._on_property_toggled)
            self.content_layout.addWidget(prop_widget)
            self.property_widgets[prop['name']] = prop_widget
        
        self._update_status()
        self._emit_selection()
    
    def _on_property_toggled(self, prop_name, checked):
        """Handle when a property checkbox is toggled."""
        self._update_status()
        self._emit_selection()
    
    def _on_variable_renderer_changed(self, variable_name, renderer_class):
        """Handle when a variable's renderer selection changes."""
        self._variable_renderers[variable_name] = renderer_class
        
        # If synchronization is enabled and we're not already syncing, update all variables with the same name
        if self.sync_checkbox.isChecked() and not self._is_syncing:
            self._sync_variable_renderer(variable_name, renderer_class)
        
        # Prevent recursive signal calls
        if not self._is_syncing:
            self.renderers_changed.emit()
    
    def _sync_variable_renderer(self, variable_name, renderer_class):
        """Update all variables with the same name to use the given renderer."""
        self._is_syncing = True  # Set flag to prevent recursive calls
        try:
            for prop_widget in self.property_widgets.values():
                var_widget = prop_widget.variable_widgets.get(variable_name)
                var_widget._set_renderer_class(renderer_class)
        except Exception as e:
            print(f"Error during variable renderer synchronization: {e}")
        finally:
            self._is_syncing = False

    def selected_properties(self):
        """Return list of selected property names"""
        return [name for name, widget in self.property_widgets.items() if widget.is_checked()]

    def select_all(self):
        self._set_all(True)

    def select_none(self):
        self._set_all(False)

    def _set_all(self, checked):
        for widget in self.property_widgets.values():
            widget.set_checked(checked)
        self._update_status()
        self._emit_selection()

    def _emit_selection(self):
        self._update_status()
        self.selection_changed.emit(self.selected_properties())

    def _update_status(self):
        total = len(self.property_widgets)
        selected = len(self.selected_properties())
        self.status_label.setText(f"{selected} / {total} selected")

    @property
    def property_items(self):
        return [(name, widget) for name, widget in self.property_widgets.items()]
    
    def get_variable_renderers(self):
        """Return a dict of property_name-variable_name -> renderer_class for all variables."""
        renderers = {}
        for prop_name, prop_widget in self.property_widgets.items():
            for var_name, var_widget in prop_widget.variable_widgets.items():
                key = f"{prop_name}-{var_name}"
                renderers[key] = var_widget.renderer_class
        return renderers


"""
This is a CLI for running Vehicle commands in a separate process. 
It is used to run the VCL commands in a separate process to avoid blocking the main thread.
It is also used to stream the output of the VCL commands to the main thread using asyncio.
This is a stop-gap until we develop a threaded solution inside of the Vehicle codebase.
"""

from vehicle_lang.session import Session
from vehicle_gui.vcl_bindings import CACHE_DIR
import sys
import os

if __name__ == "__main__":
	# Get the command line arguments
	args = sys.argv[1:]

	# Create a new session
	s = Session().__enter__()

	# Run the command and get the output
	log_path = os.path.join(os.path.dirname(__file__), CACHE_DIR, "log.txt")
	try:
		s.check_call(
			[
				f"--redirect-logs={log_path}",
				*args,
			]
		)
	finally:
		if os.path.exists(log_path):
			os.remove(log_path)



"""
Module palette.py

This module contains the color style references for the graphics

Original author: Andrea Gimelli, Giacomo Rosato, Stefano Demarchi

"""

DARK_TEAL = "#105555"
TEAL = "#157172"
LIGHT_TEAL = "#1ABC9C"

GREEN_1 = "#0A6160"
GREEN_2 = "#0C8371"
LIGHT_GREEN = "#27AE60"
DARK_GREEN = "#1E8449"

RED = "#b80000"
RED_1 = "#FF0055"
RED_2 = "#E60000"
LIGHT_RED = "#E74C3C"
DARK_RED = "#C0392B"

DARK_BLUE = "#304153"
BLUE = "#34475b"
BLUE_0 = "#005588"
BLUE_1 = "#0077AA"

DARK_GREY = "#2c2c2c"
LESS_DARK_GREY = "#303030"
GREY = "#4B4E52"
LIGHT_GREY = "#56595d"
GREY_0 = "#2A2A2A"
GREY_1 = "#323232"
GREY_2 = "#4D4D4D"
GREY_3 = "#656565"
GREY_4 = "#6F6F6F"

RED = "#b80000"
RED_1 = "#FF0055"
RED_2 = "#E60000"

DARK_ORANGE = "#873511"
ORANGE = "#943A12"
ORANGE_0 = "#FF9900"
ORANGE_1 = "#FFAD33"

BACKGROUND_GREY = "#393939"
BACKGROUND_LIGHT_LINE_GREY = "#2f2f2f"
BACKGROUND_DARK_LINE_GREY = "#292929"

BLACK = "black"
WHITE = "white"
WHITE_GRAY = "#F2F2F2"

VIOLET   = "#7C3AED"
VIOLET_1 = "#8B5CF6"



import os
import sys


def which_all(cmd, mode=os.F_OK | os.X_OK, path=None):
    """Return a list of full paths to all executables named `cmd` on PATH."""
    if path is None:
        path = os.environ.get("PATH", os.defpath)
    paths = path.split(os.pathsep)

    # If cmd has a path separator, check it directly.
    if os.path.dirname(cmd):
        p = cmd
        return [p] if os.path.isfile(p) and os.access(p, mode) else []

    # Windows: respect PATHEXT; POSIX: just the name.
    exts = []
    if sys.platform == "win32":
        pathext = os.environ.get("PATHEXT", ".COM;.EXE;.BAT;.CMD")
        exts = [e.lower() for e in pathext.split(os.pathsep)]
        if any(cmd.lower().endswith(e) for e in exts):
            exts = [""]

    candidates = [cmd + ext for ext in (exts or [""])]

    found, seen = [], set()
    for d in paths:
        d = d.strip('"')
        if not d:
            continue
        for name in candidates:
            full = os.path.join(d, name)
            if os.path.isfile(full) and os.access(full, mode):
                # Deduplicate by real path to avoid repeats via symlinks.
                rp = os.path.realpath(full)
                if rp not in seen:
                    seen.add(rp)
                    found.append(full)
    return found


"""
Module dimension.py

This module contains the dimension references for the graphics

Original author: Andrea Gimelli, Giacomo Rosato, Stefano Demarchi

"""

# Scene
# ------------

# Minimum scene dimensions
INITIAL_SCENE_WIDTH = 800
INITIAL_SCENE_HEIGHT = 600
SCENE_PADDING = 200

# Window size
WINDOW_WIDTH = 1024
WINDOW_HEIGHT = 768

# Other dimensions
GRID_SIZE = 20
GRID_SQUARE = 5
ABS_BLOCK_DISTANCE = 100

# Graphics View
# ------------

# Zoom parameters
ZOOM = 10
ZOOM_IN_FACTOR = 1.25
ZOOM_STEP = 1
ZOOM_RANGE = [0, 10]

# Blocks
# ------------

# Width values
BLOCK_BASE_WIDTH = 120

# Height values
BLOCK_BASE_HEIGHT = 35
TITLE_HEIGHT = 25

# Other dimensions
EDGE_ROUNDNESS = 10
EDGE_CP_ROUNDNESS = 100
TITLE_PAD = 10.0
SOCKET_SPACING = 22

NEXT_BLOCK_DISTANCE = 200
SCENE_BORDER = 4600

# Font
FONT_FAMILY = "Arial"
FONT_SIZE = 8
FONT_BOLD = True

# Sockets
# ------------

# Parameters
SOCKET_RADIUS = 2
SOCKET_OUTLINE = 1

# Verification Workflow Layout
# ------------

# Hierarchical tree layout spacing
VERTICAL_SPACING = 100      # Vertical spacing between blocks
HORIZONTAL_SPACING = 50    # Horizontal spacing between blocks
PROPERTY_SPACING_Y = 300    # Vertical spacing between properties
PROPERTY_X = -400           # X position for properties

# Starting position for workflow
WORKFLOW_START_Y = -300     # Starting Y position for first property



"""
Module blocks.py

This module contains verification-specific block types for the node editor.

"""

from enum import Enum
from ..base_types import Block
from ..styling.palette import *


PALETTE = {
    # [_pen_default, _pen_hovered, _pen_selected, _brush_title, _brush_background]
    "AND":       ( [BLUE_0, BLUE_1, DARK_BLUE, BLUE_0, DARK_GREY],                  {"bg_color": BLUE_1,    "outline_color": BLUE_0} ),
    "OR":        ( [VIOLET, VIOLET_1, BLUE, VIOLET, DARK_GREY],                     {"bg_color": VIOLET_1,  "outline_color": VIOLET} ),
    "VERIFIED":  ( [DARK_GREEN, LIGHT_GREEN, LIGHT_GREEN, DARK_GREEN, DARK_GREY],   {"bg_color": LIGHT_GREEN, "outline_color": DARK_GREEN} ),
    "DISPROVEN": ( [DARK_RED, LIGHT_RED, LIGHT_RED, DARK_RED, DARK_GREY],           {"bg_color": LIGHT_RED,   "outline_color": DARK_RED} ),
    "UNKNOWN":   ( [DARK_ORANGE, ORANGE_1, ORANGE_0, DARK_ORANGE, DARK_GREY],       {"bg_color": ORANGE_1,    "outline_color": DARK_ORANGE} ),
    "WITNESS":   ( [DARK_TEAL, TEAL, LIGHT_TEAL, DARK_TEAL, DARK_GREY],             {"bg_color": TEAL,        "outline_color": DARK_TEAL} ),
}


class BlockType(Enum):
    """Block type enumeration"""
    PROPERTY = 0
    QUERY = 1
    WITNESS = 2


class Status(Enum):
    """Verification status enumeration"""
    VERIFIED = 0
    DISPROVEN = 1
    UNKNOWN = 2


class PropertyQuantifier(Enum):
    """Property quantifier enumeration"""
    FOR_ALL = 0
    EXISTS = 1


class OrBlock(Block):
    """Block representing a disjunction (OR) in verification properties"""
    def __init__(self, parent=None):
        super().__init__()
        self.title = "OR"
        self.children = []  # List of child blocks
        self.parent_ref = parent  # Reference to parent block
        self.verification_status = Status.UNKNOWN   # Whether this proves or disproves the original property
    
    def get_color_scheme(self):
        """Return color scheme for OR blocks"""
        return PALETTE["OR"][0]
    
    def get_socket_colors(self):
        """Return socket colors for OR blocks"""
        return PALETTE["OR"][1]


class AndBlock(Block):
    """Block representing a conjunction (AND) in verification properties"""
    def __init__(self, parent=None):
        super().__init__()
        self.title = "AND"
        self.children = []  # List of child blocks
        self.parent_ref = parent  # Reference to parent block
        self.verification_status = Status.UNKNOWN       # Whether this proves or disproves the original property

    def get_color_scheme(self):
        """Return color scheme for AND blocks"""
        # [_pen_default, _pen_hovered , _pen_selected, _brush_title, _brush_background]
        return PALETTE["AND"][0]

    def get_socket_colors(self):
        """Return socket colors for AND blocks"""
        return PALETTE["AND"][1]


class PropertyBlock(Block):
    """Block representing a high-level verification property"""
    def __init__(self, title="Property"):
        super().__init__()
        self.title = title
        self.verification_status = Status.UNKNOWN
        self.children = []  # List of child blocks
        self.queries = []   # List of associated queries

    def get_color_scheme(self):
        """Return color scheme based on verification status"""
        key = "UNKNOWN"
        if self.verification_status == Status.VERIFIED:
            key = "VERIFIED"
        elif self.verification_status == Status.DISPROVEN:
            key = "DISPROVEN"
        return PALETTE[key][0]
    
    def get_socket_colors(self):
        """Return socket colors for property blocks"""
        key = "UNKNOWN"
        if self.verification_status == Status.VERIFIED:
            key = "VERIFIED"
        elif self.verification_status == Status.DISPROVEN:
            key = "DISPROVEN"
        return PALETTE[key][1]


class QueryBlock(Block):
    """Block representing a verification query"""
    def __init__(self, id, parent, path, is_negated=False):
        super().__init__()
        self.id = id
        self.parent_ref = parent
        self.path = path
        self.title = f"Query {id}" if not is_negated else f"¬Query {id}"
        self.is_negated = is_negated 
        self.verification_status = Status.UNKNOWN
    
    def get_color_scheme(self):
        """Return color scheme based on verification status"""
        key = "UNKNOWN"
        if self.verification_status == Status.VERIFIED:
            key = "VERIFIED"
        elif self.verification_status == Status.DISPROVEN:
            key = "DISPROVEN"
        return PALETTE[key][0]
    
    def get_socket_colors(self):
        """Return socket colors for query blocks"""
        key = "UNKNOWN"
        if self.verification_status == Status.VERIFIED:
            key = "VERIFIED"
        elif self.verification_status == Status.DISPROVEN:
            key = "DISPROVEN"
        return PALETTE[key][1]


class WitnessBlock(Block):
    """Block representing verification results (witness/counterexample)"""
    def __init__(self, query_block):
        super().__init__()
        self.is_counterexample = query_block.is_negated
        self.query_ref = query_block  # Reference to associated query
        self.title = "Counter Example" if self.is_counterexample else "Witness"

    def get_color_scheme(self):
        """Return color scheme based on counterexample vs witness"""
        return PALETTE["DISPROVEN"][0] if self.is_counterexample else PALETTE["WITNESS"][0]
    
    def get_socket_colors(self):
        """Return socket colors for witness blocks"""
        return PALETTE["DISPROVEN"][1] if self.is_counterexample else PALETTE["WITNESS"][1]
